# Copyright (c) 2025 Contributors to the Eclipse Foundation
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Apache Software License 2.0 which is available at
# https://www.apache.org/licenses/LICENSE-2.0, or the MIT license
# which is available at https://opensource.org/licenses/MIT.
#
# SPDX-License-Identifier: Apache-2.0 OR MIT

import ctypes

import iceoryx2 as iox2
import pytest

service_types = [iox2.ServiceType.Ipc, iox2.ServiceType.Local]


class Payload(ctypes.Structure):
    _fields_ = [("data", ctypes.c_ubyte)]


class LargePayload(ctypes.Structure):
    _fields_ = [("data", ctypes.c_ulonglong)]


@pytest.mark.parametrize("service_type", service_types)
def test_send_and_receive_with_memmove_works(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)
    number_of_samples = 5

    service_name = iox2.testing.generate_service_name()
    service = (
        node.service_builder(service_name)
        .publish_subscribe(Payload)
        .subscriber_max_buffer_size(number_of_samples)
        .create()
    )

    publisher = service.publisher_builder().create()
    subscriber = service.subscriber_builder().create()
    assert not subscriber.has_samples()

    for i in range(0, number_of_samples):
        send_payload = Payload(data=82 + i)
        sample_uninit = publisher.loan_uninit()
        ctypes.memmove(
            sample_uninit.payload_ptr,
            ctypes.byref(send_payload),
            ctypes.sizeof(Payload),
        )
        sample = sample_uninit.assume_init()
        sample.send()

    assert subscriber.has_samples()

    for i in range(0, number_of_samples):
        assert subscriber.has_samples()
        received_sample = subscriber.receive()
        assert received_sample is not None
        received_payload = Payload(data=0)
        ctypes.memmove(
            ctypes.byref(received_payload),
            received_sample.payload_ptr,
            ctypes.sizeof(Payload),
        )
        assert received_payload.data == 82 + i

    assert not subscriber.has_samples()


@pytest.mark.parametrize("service_type", service_types)
def test_send_copy_and_receive_works(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)
    number_of_samples = 6

    service_name = iox2.testing.generate_service_name()
    service = (
        node.service_builder(service_name)
        .publish_subscribe(Payload)
        .subscriber_max_buffer_size(number_of_samples)
        .create()
    )

    publisher = service.publisher_builder().create()
    subscriber = service.subscriber_builder().create()
    assert not subscriber.has_samples()

    for i in range(0, number_of_samples):
        publisher.send_copy(Payload(data=85 + i))

    assert subscriber.has_samples()

    for i in range(0, number_of_samples):
        received_sample = subscriber.receive()
        assert received_sample.payload().contents.data == 85 + i

    assert not subscriber.has_samples()


@pytest.mark.parametrize("service_type", service_types)
def test_send_with_write_payload_and_receive_works(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)
    number_of_samples = 6

    service_name = iox2.testing.generate_service_name()
    service = (
        node.service_builder(service_name)
        .publish_subscribe(Payload)
        .subscriber_max_buffer_size(number_of_samples)
        .create()
    )

    publisher = service.publisher_builder().create()
    subscriber = service.subscriber_builder().create()
    assert not subscriber.has_samples()

    for i in range(0, number_of_samples):
        sample_uninit = publisher.loan_uninit()
        sample = sample_uninit.write_payload(Payload(data=89 + i))
        sample.send()

    assert subscriber.has_samples()

    for i in range(0, number_of_samples):
        received_sample = subscriber.receive()
        assert received_sample.payload().contents.data == 89 + i

    assert not subscriber.has_samples()


@pytest.mark.parametrize("service_type", service_types)
def test_send_large_payload_works(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    service = (
        node.service_builder(service_name).publish_subscribe(LargePayload).create()
    )

    publisher = service.publisher_builder().create()
    subscriber = service.subscriber_builder().create()

    send_payload = LargePayload(data=19203182930990147)
    sample_uninit = publisher.loan_uninit()
    ctypes.memmove(sample_uninit.payload_ptr, ctypes.byref(send_payload), 8)
    sample = sample_uninit.assume_init()
    sample.send()

    received_sample = subscriber.receive()
    assert received_sample is not None
    received_payload = LargePayload(data=0)
    ctypes.memmove(ctypes.byref(received_payload), received_sample.payload_ptr, 8)
    assert received_payload.data == send_payload.data


@pytest.mark.parametrize("service_type", service_types)
def test_published_header_is_the_same_as_received_header(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    service = node.service_builder(service_name).publish_subscribe(Payload).create()

    publisher = service.publisher_builder().create()
    subscriber = service.subscriber_builder().create()

    sample_uninit = publisher.loan_uninit()
    sample = sample_uninit.assume_init()
    send_header = sample.header
    assert send_header.node_id == node.id
    assert send_header.publisher_id == publisher.id
    assert send_header.number_of_elements == 1

    sample.send()

    received_sample = subscriber.receive()
    assert received_sample is not None
    assert received_sample.header == send_header


@pytest.mark.parametrize("service_type", service_types)
def test_custom_user_header_can_be_used(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    service = (
        node.service_builder(service_name)
        .publish_subscribe(Payload)
        .user_header(Payload)
        .create()
    )

    publisher = service.publisher_builder().create()
    subscriber = service.subscriber_builder().create()

    sample_uninit = publisher.loan_uninit()
    send_user_header_payload = Payload(data=37)
    ctypes.memmove(
        sample_uninit.user_header_ptr, ctypes.byref(send_user_header_payload), 1
    )
    sample = sample_uninit.assume_init()
    sample.send()

    received_sample = subscriber.receive()
    assert received_sample is not None
    assert received_sample.user_header().contents.data == send_user_header_payload.data


@pytest.mark.parametrize("service_type", service_types)
def test_reallocation_fails_when_allocation_strategy_is_static(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    service = (
        node.service_builder(service_name)
        .publish_subscribe(iox2.Slice[ctypes.c_uint8])
        .create()
    )

    publisher = (
        service.publisher_builder()
        .initial_max_slice_len(8)
        .allocation_strategy(iox2.AllocationStrategy.Static)
        .create()
    )

    try:
        publisher.loan_slice_uninit(8)
    except iox2.LoanError:
        assert False

    with pytest.raises(iox2.LoanError):
        publisher.loan_slice_uninit(9)


@pytest.mark.parametrize("service_type", service_types)
def test_reallocation_works_when_allocation_strategy_is_not_static(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    service = (
        node.service_builder(service_name)
        .publish_subscribe(iox2.Slice[ctypes.c_uint8])
        .create()
    )

    publisher = (
        service.publisher_builder()
        .initial_max_slice_len(8)
        .allocation_strategy(iox2.AllocationStrategy.PowerOfTwo)
        .create()
    )

    try:
        publisher.loan_slice_uninit(12)
    except iox2.LoanError:
        assert False


@pytest.mark.parametrize("service_type", service_types)
def test_slice_type_forbids_use_of_non_slice_api(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    service = (
        node.service_builder(service_name)
        .publish_subscribe(iox2.Slice[ctypes.c_uint8])
        .create()
    )

    publisher = service.publisher_builder().create()

    with pytest.raises(AssertionError):
        publisher.loan_uninit()


@pytest.mark.parametrize("service_type", service_types)
def test_non_slice_type_forbids_use_of_slice_api(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    service = node.service_builder(service_name).publish_subscribe(Payload).create()

    with pytest.raises(AssertionError):
        publisher = service.publisher_builder().initial_max_slice_len(8).create()

    with pytest.raises(AssertionError):
        publisher = (
            service.publisher_builder()
            .allocation_strategy(iox2.AllocationStrategy.PowerOfTwo)
            .create()
        )

    publisher = service.publisher_builder().create()

    with pytest.raises(AssertionError):
        publisher.loan_slice_uninit(1)



# Copyright (c) 2025 Contributors to the Eclipse Foundation
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Apache Software License 2.0 which is available at
# https://www.apache.org/licenses/LICENSE-2.0, or the MIT license
# which is available at https://opensource.org/licenses/MIT.
#
# SPDX-License-Identifier: Apache-2.0 OR MIT

import ctypes

import iceoryx2 as iox2
import pytest

service_types = [iox2.ServiceType.Ipc, iox2.ServiceType.Local]


class Payload(ctypes.Structure):
    _fields_ = [("data", ctypes.c_ubyte)]


class HeaderPayload(ctypes.Structure):
    _fields_ = [("data", ctypes.c_ubyte), ("fuu", ctypes.c_int)]


class Payload2(ctypes.Structure):
    _fields_ = [
        ("data", ctypes.c_uint64),
        ("fuu", ctypes.c_uint64),
        ("bar", ctypes.c_uint64),
    ]


class HeaderPayload2(ctypes.Structure):
    _fields_ = [
        ("data", ctypes.c_uint64),
        ("fuu", ctypes.c_uint64),
        ("bar", ctypes.c_uint64),
        ("whatever", ctypes.c_uint64),
    ]


@pytest.mark.parametrize("service_type", service_types)
def test_non_existing_service_can_be_created(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)
    try:
        service_name = iox2.testing.generate_service_name()
        sut = (
            node.service_builder(service_name)
            .request_response(Payload, Payload)
            .create()
        )
        assert sut.name == service_name
    except iox2.RequestResponseCreateError:
        assert False


@pytest.mark.parametrize("service_type", service_types)
def test_existing_service_cannot_be_created(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    _existing_service = (
        node.service_builder(service_name).request_response(Payload, Payload).create()
    )

    with pytest.raises(iox2.RequestResponseCreateError):
        node.service_builder(service_name).request_response(Payload, Payload).create()


@pytest.mark.parametrize("service_type", service_types)
def test_existing_service_can_be_opened(service_type: iox2.ServiceType) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    _existing_service = (
        node.service_builder(service_name).request_response(Payload, Payload).create()
    )
    try:
        sut = (
            node.service_builder(service_name).request_response(Payload, Payload).open()
        )
        assert sut.name == service_name
    except iox2.RequestResponseOpenError:
        assert False


@pytest.mark.parametrize("service_type", service_types)
def test_non_existing_service_cannot_be_opened(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    with pytest.raises(iox2.RequestResponseOpenError):
        node.service_builder(service_name).request_response(Payload, Payload).open()


@pytest.mark.parametrize("service_type", service_types)
def test_non_existing_service_is_created_with_open_or_create(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    try:
        sut = (
            node.service_builder(service_name)
            .request_response(Payload, Payload)
            .open_or_create()
        )
        assert sut.name == service_name
    except iox2.RequestResponseOpenOrCreateError:
        assert False


@pytest.mark.parametrize("service_type", service_types)
def test_existing_service_is_opened_with_open_or_create(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    _existing_service = (
        node.service_builder(service_name).request_response(Payload, Payload).create()
    )

    try:
        sut = (
            node.service_builder(service_name)
            .request_response(Payload, Payload)
            .open_or_create()
        )
        assert sut.name == service_name
    except iox2.RequestResponseOpenOrCreateError:
        assert False


@pytest.mark.parametrize("service_type", service_types)
def test_create_service_with_attributes_work(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)
    attribute_spec = iox2.AttributeSpecifier.new().define(
        iox2.AttributeKey.new("fuu"), iox2.AttributeValue.new("bar")
    )

    service_name = iox2.testing.generate_service_name()
    sut_create = (
        node.service_builder(service_name)
        .request_response(Payload, Payload)
        .create_with_attributes(attribute_spec)
    )

    sut_open = (
        node.service_builder(service_name).request_response(Payload, Payload).open()
    )

    assert sut_create.attributes == attribute_spec.attributes
    assert sut_create.attributes == sut_open.attributes


@pytest.mark.parametrize("service_type", service_types)
def test_open_or_create_service_with_attributes_work(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    attribute_spec = iox2.AttributeSpecifier.new().define(
        iox2.AttributeKey.new("what"), iox2.AttributeValue.new("ever")
    )
    attribute_verifier = iox2.AttributeVerifier.new().require(
        iox2.AttributeKey.new("what"), iox2.AttributeValue.new("ever")
    )

    service_name = iox2.testing.generate_service_name()
    sut_create = (
        node.service_builder(service_name)
        .request_response(Payload, Payload)
        .open_or_create_with_attributes(attribute_verifier)
    )

    sut_open = (
        node.service_builder(service_name).request_response(Payload, Payload).open()
    )

    assert sut_create.attributes == attribute_spec.attributes
    assert sut_create.attributes == sut_open.attributes


@pytest.mark.parametrize("service_type", service_types)
def test_open_service_with_attributes_work(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    attribute_spec = iox2.AttributeSpecifier.new().define(
        iox2.AttributeKey.new("knock"), iox2.AttributeValue.new("knock")
    )
    attribute_verifier = iox2.AttributeVerifier.new().require(
        iox2.AttributeKey.new("knock"), iox2.AttributeValue.new("knock")
    )

    service_name = iox2.testing.generate_service_name()
    sut_create = (
        node.service_builder(service_name)
        .request_response(Payload, Payload)
        .create_with_attributes(attribute_spec)
    )

    sut_open = (
        node.service_builder(service_name)
        .request_response(Payload, Payload)
        .open_with_attributes(attribute_verifier)
    )

    assert sut_create.attributes == attribute_spec.attributes
    assert sut_create.attributes == sut_open.attributes


@pytest.mark.parametrize("service_type", service_types)
def test_node_listing_works(service_type: iox2.ServiceType) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    sut = node.service_builder(service_name).request_response(Payload, Payload).create()

    nodes = sut.nodes

    assert len(nodes) == 1
    for n in nodes:
        match n:
            case n.Alive():
                assert n[0].id == node.id
            case node.Dead():
                assert False
            case node.Inaccessible():
                assert False
            case node.Undefined():
                assert False


@pytest.mark.parametrize("service_type", service_types)
def test_service_builder_configuration_works(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    safe_overflow_requests = False
    safe_overflow_responses = False
    fire_and_forget = False
    max_active_requests_per_client = 99
    max_loaned_requests = 88
    max_response_buffer_size = 77
    max_servers = 66
    max_clients = 55
    max_nodes = 44
    max_borrowed_responses_per_pending_response = 33
    sut = (
        node.service_builder(service_name)
        .request_response(Payload, Payload)
        .enable_safe_overflow_for_requests(safe_overflow_requests)
        .enable_safe_overflow_for_responses(safe_overflow_responses)
        .enable_fire_and_forget_requests(fire_and_forget)
        .max_active_requests_per_client(max_active_requests_per_client)
        .max_loaned_requests(max_loaned_requests)
        .max_response_buffer_size(max_response_buffer_size)
        .max_servers(max_servers)
        .max_clients(max_clients)
        .max_nodes(max_nodes)
        .max_borrowed_responses_per_pending_response(
            max_borrowed_responses_per_pending_response
        )
        .create()
    )

    static_config = sut.static_config
    assert static_config.has_safe_overflow_for_requests == safe_overflow_requests
    assert static_config.has_safe_overflow_for_responses == safe_overflow_responses
    assert static_config.does_support_fire_and_forget_requests == fire_and_forget
    assert (
        static_config.max_borrowed_responses_per_pending_response
        == max_borrowed_responses_per_pending_response
    )
    assert (
        static_config.max_active_requests_per_client == max_active_requests_per_client
    )
    assert static_config.max_response_buffer_size == max_response_buffer_size
    assert static_config.max_loaned_requests == max_loaned_requests
    assert static_config.max_servers == max_servers
    assert static_config.max_clients == max_clients
    assert static_config.max_nodes == max_nodes


@pytest.mark.parametrize("service_type", service_types)
def test_custom_request_payload_works(service_type: iox2.ServiceType) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    payload = (
        iox2.TypeDetail.new()
        .type_variant(iox2.TypeVariant.FixedSize)
        .type_name(iox2.TypeName.new("Payload"))
        .size(ctypes.sizeof(Payload))
        .alignment(ctypes.alignment(Payload))
    )
    sut = (
        node.service_builder(service_name).request_response(Payload, Payload2).create()
    )

    assert sut.static_config.request_message_type_details.payload == payload


@pytest.mark.parametrize("service_type", service_types)
def test_custom_request_header_works(service_type: iox2.ServiceType) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    user_header = (
        iox2.TypeDetail.new()
        .type_variant(iox2.TypeVariant.FixedSize)
        .type_name(iox2.TypeName.new("HeaderPayload"))
        .size(ctypes.sizeof(HeaderPayload))
        .alignment(ctypes.alignment(HeaderPayload))
    )
    sut = (
        node.service_builder(service_name)
        .request_response(Payload, Payload)
        .request_header(HeaderPayload)
        .create()
    )

    assert sut.static_config.request_message_type_details.user_header == user_header


@pytest.mark.parametrize("service_type", service_types)
def test_custom_response_payload_works(service_type: iox2.ServiceType) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    payload = (
        iox2.TypeDetail.new()
        .type_variant(iox2.TypeVariant.FixedSize)
        .type_name(iox2.TypeName.new("Payload2"))
        .size(ctypes.sizeof(Payload2))
        .alignment(ctypes.alignment(Payload2))
    )
    sut = (
        node.service_builder(service_name).request_response(Payload, Payload2).create()
    )

    assert sut.static_config.response_message_type_details.payload == payload


@pytest.mark.parametrize("service_type", service_types)
def test_custom_response_header_works(service_type: iox2.ServiceType) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    user_header = (
        iox2.TypeDetail.new()
        .type_variant(iox2.TypeVariant.FixedSize)
        .type_name(iox2.TypeName.new("HeaderPayload2"))
        .size(ctypes.sizeof(HeaderPayload2))
        .alignment(ctypes.alignment(HeaderPayload2))
    )
    sut = (
        node.service_builder(service_name)
        .request_response(Payload, Payload)
        .response_header(HeaderPayload2)
        .create()
    )

    assert sut.static_config.response_message_type_details.user_header == user_header



# Copyright (c) 2025 Contributors to the Eclipse Foundation
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Apache Software License 2.0 which is available at
# https://www.apache.org/licenses/LICENSE-2.0, or the MIT license
# which is available at https://opensource.org/licenses/MIT.
#
# SPDX-License-Identifier: Apache-2.0 OR MIT

import iceoryx2 as iox2
import pytest


def test_attribute_verifier_require_works() -> None:
    key = iox2.AttributeKey.new("ding dong")
    value = iox2.AttributeValue.new("rainbow")

    sut = iox2.AttributeVerifier.new().require(key, value)
    specifier = iox2.AttributeSpecifier.new().define(key, value)

    assert sut.required_attributes == specifier.attributes
    assert sut.required_attributes.number_of_attributes == 1
    attribute = sut.required_attributes.values
    assert attribute[0].key == key
    assert attribute[0].value == value


def test_attribute_verifier_require_key_works() -> None:
    key = iox2.AttributeKey.new("me and my brain")
    value = iox2.AttributeValue.new("we are not friends")

    sut = iox2.AttributeVerifier.new().require_key(key)
    specifier = iox2.AttributeSpecifier.new().define(key, value)

    keys = sut.required_keys
    assert len(keys) == 1
    assert keys[0] == key

    assert sut.verify_requirements(specifier.attributes) is None



# Copyright (c) 2025 Contributors to the Eclipse Foundation
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Apache Software License 2.0 which is available at
# https://www.apache.org/licenses/LICENSE-2.0, or the MIT license
# which is available at https://opensource.org/licenses/MIT.
#
# SPDX-License-Identifier: Apache-2.0 OR MIT

import iceoryx2 as iox2
import pytest


def test_duration_from_secs_works() -> None:
    sut = iox2.Duration.from_secs(2)
    assert sut.as_secs() == 2


def test_duration_from_millis_works() -> None:
    sut = iox2.Duration.from_millis(25)
    assert sut.as_millis() == 25


def test_duration_from_micros_works() -> None:
    sut = iox2.Duration.from_micros(981)
    assert sut.as_micros() == 981


def test_duration_from_nanos_works() -> None:
    sut = iox2.Duration.from_nanos(421)
    assert sut.as_nanos() == 421


def test_duration_from_secs_f64_works() -> None:
    sut = iox2.Duration.from_secs_f64(12.34)
    assert sut.as_secs() == 12
    assert sut.as_millis() == 12340
    assert sut.as_secs_f64() == 12.34


def test_duration_subsec_micros_works() -> None:
    sut = iox2.Duration.from_secs_f64(2.34567891011)
    assert sut.subsec_micros() == 345678


def test_duration_subsec_millis_works() -> None:
    sut = iox2.Duration.from_secs_f64(3.4567890101112)
    assert sut.subsec_millis() == 456


def test_duration_subsec_nanos_works() -> None:
    sut = iox2.Duration.from_secs_f64(4.567891011)
    assert sut.subsec_nanos() == 567891011



# Copyright (c) 2025 Contributors to the Eclipse Foundation
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Apache Software License 2.0 which is available at
# https://www.apache.org/licenses/LICENSE-2.0, or the MIT license
# which is available at https://opensource.org/licenses/MIT.
#
# SPDX-License-Identifier: Apache-2.0 OR MIT

import ctypes

import iceoryx2 as iox2
import pytest

service_types = [iox2.ServiceType.Ipc, iox2.ServiceType.Local]


class Payload(ctypes.Structure):
    _fields_ = [
        ("data", ctypes.c_int),
    ]


@pytest.mark.parametrize("service_type", service_types)
def test_can_be_configured(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)
    service_name = iox2.testing.generate_service_name()
    service = (
        node.service_builder(service_name)
        .publish_subscribe(Payload)
        .subscriber_max_buffer_size(47112)
        .create()
    )

    sut = service.subscriber_builder().buffer_size(47112).create()

    assert sut.buffer_size == 47112


@pytest.mark.parametrize("service_type", service_types)
def test_deleting_subscriber_removes_if_from_the_service(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)
    service_name = iox2.testing.generate_service_name()
    service = (
        node.service_builder(service_name)
        .publish_subscribe(Payload)
        .max_subscribers(1)
        .create()
    )

    sut = service.subscriber_builder().create()

    with pytest.raises(iox2.SubscriberCreateError):
        sut = service.subscriber_builder().create()

    sut.delete()

    try:
        sut = service.subscriber_builder().create()
    except iox2.SubscriberCreateError:
        assert False


@pytest.mark.parametrize("service_type", service_types)
def test_deleting_sample_releases_it(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)
    service_name = iox2.testing.generate_service_name()
    service = (
        node.service_builder(service_name)
        .publish_subscribe(Payload)
        .subscriber_max_buffer_size(2)
        .subscriber_max_borrowed_samples(1)
        .create()
    )

    sut = service.subscriber_builder().create()
    publisher = service.publisher_builder().create()
    publisher.loan_uninit().assume_init().send()
    publisher.loan_uninit().assume_init().send()

    sample = sut.receive()
    with pytest.raises(iox2.ReceiveError):
        sample = sut.receive()

    sample.delete()

    try:
        sample = sut.receive()
    except iox2.ReceiveError:
        assert False



# Copyright (c) 2025 Contributors to the Eclipse Foundation
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Apache Software License 2.0 which is available at
# https://www.apache.org/licenses/LICENSE-2.0, or the MIT license
# which is available at https://opensource.org/licenses/MIT.
#
# SPDX-License-Identifier: Apache-2.0 OR MIT

import iceoryx2 as iox2
import pytest

service_types = [iox2.ServiceType.Ipc, iox2.ServiceType.Local]


@pytest.mark.parametrize("service_type", service_types)
def test_non_existing_service_can_be_created(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)
    try:
        service_name = iox2.testing.generate_service_name()
        sut = node.service_builder(service_name).event().create()
        assert sut.name == service_name
    except iox2.EventCreateError:
        assert False


@pytest.mark.parametrize("service_type", service_types)
def test_existing_service_cannot_be_created(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    _existing_service = node.service_builder(service_name).event().create()

    with pytest.raises(iox2.EventCreateError):
        node.service_builder(service_name).event().create()


@pytest.mark.parametrize("service_type", service_types)
def test_existing_service_can_be_opened(service_type: iox2.ServiceType) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    _existing_service = node.service_builder(service_name).event().create()
    try:
        sut = node.service_builder(service_name).event().open()
        assert sut.name == service_name
    except iox2.EventOpenError:
        assert False


@pytest.mark.parametrize("service_type", service_types)
def test_non_existing_service_cannot_be_opened(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    with pytest.raises(iox2.EventOpenError):
        node.service_builder(service_name).event().open()


@pytest.mark.parametrize("service_type", service_types)
def test_non_existing_service_is_created_with_open_or_create(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    try:
        sut = node.service_builder(service_name).event().open_or_create()
        assert sut.name == service_name
    except iox2.EventOpenOrCreateError:
        assert False


@pytest.mark.parametrize("service_type", service_types)
def test_existing_service_is_opened_with_open_or_create(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    _existing_service = node.service_builder(service_name).event().create()

    try:
        sut = node.service_builder(service_name).event().open_or_create()
        assert sut.name == service_name
    except iox2.EventOpenOrCreateError:
        assert False


@pytest.mark.parametrize("service_type", service_types)
def test_create_service_with_attributes_work(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)
    attribute_spec = iox2.AttributeSpecifier.new().define(
        iox2.AttributeKey.new("fuu"), iox2.AttributeValue.new("bar")
    )

    service_name = iox2.testing.generate_service_name()
    sut_create = (
        node.service_builder(service_name)
        .event()
        .create_with_attributes(attribute_spec)
    )

    sut_open = node.service_builder(service_name).event().open()

    assert sut_create.attributes == attribute_spec.attributes
    assert sut_create.attributes == sut_open.attributes


@pytest.mark.parametrize("service_type", service_types)
def test_open_or_create_service_with_attributes_work(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    attribute_spec = iox2.AttributeSpecifier.new().define(
        iox2.AttributeKey.new("what"), iox2.AttributeValue.new("ever")
    )
    attribute_verifier = iox2.AttributeVerifier.new().require(
        iox2.AttributeKey.new("what"), iox2.AttributeValue.new("ever")
    )

    service_name = iox2.testing.generate_service_name()
    sut_create = (
        node.service_builder(service_name)
        .event()
        .open_or_create_with_attributes(attribute_verifier)
    )

    sut_open = node.service_builder(service_name).event().open()

    assert sut_create.attributes == attribute_spec.attributes
    assert sut_create.attributes == sut_open.attributes


@pytest.mark.parametrize("service_type", service_types)
def test_open_service_with_attributes_work(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    attribute_spec = iox2.AttributeSpecifier.new().define(
        iox2.AttributeKey.new("knock"), iox2.AttributeValue.new("knock")
    )
    attribute_verifier = iox2.AttributeVerifier.new().require(
        iox2.AttributeKey.new("knock"), iox2.AttributeValue.new("knock")
    )

    service_name = iox2.testing.generate_service_name()
    sut_create = (
        node.service_builder(service_name)
        .event()
        .create_with_attributes(attribute_spec)
    )

    sut_open = (
        node.service_builder(service_name)
        .event()
        .open_with_attributes(attribute_verifier)
    )

    assert sut_create.attributes == attribute_spec.attributes
    assert sut_create.attributes == sut_open.attributes


@pytest.mark.parametrize("service_type", service_types)
def test_node_listing_works(service_type: iox2.ServiceType) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    sut = node.service_builder(service_name).event().create()

    nodes = sut.nodes

    assert len(nodes) == 1
    for n in nodes:
        match n:
            case n.Alive():
                assert n[0].id == node.id
            case node.Dead():
                assert False
            case node.Inaccessible():
                assert False
            case node.Undefined():
                assert False


@pytest.mark.parametrize("service_type", service_types)
def test_service_builder_configuration_works(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    deadline = iox2.Duration.from_millis(123)
    max_nodes = 456
    event_id_max = 78
    max_notifiers = 9
    max_listeners = 10
    notifier_created = iox2.EventId.new(321)
    notifier_dead = iox2.EventId.new(654)
    notifier_dropped = iox2.EventId.new(987)
    sut = (
        node.service_builder(service_name)
        .event()
        .deadline(deadline)
        .max_nodes(max_nodes)
        .event_id_max_value(event_id_max)
        .max_notifiers(max_notifiers)
        .max_listeners(max_listeners)
        .notifier_created_event(notifier_created)
        .notifier_dropped_event(notifier_dropped)
        .notifier_dead_event(notifier_dead)
        .create()
    )

    static_config = sut.static_config
    assert static_config.deadline == deadline
    assert static_config.max_nodes == max_nodes
    assert static_config.max_listeners == max_listeners
    assert static_config.max_notifiers == max_notifiers
    assert static_config.event_id_max_value == event_id_max
    assert static_config.notifier_created_event == notifier_created
    assert static_config.notifier_dead_event == notifier_dead
    assert static_config.notifier_dropped_event == notifier_dropped


@pytest.mark.parametrize("service_type", service_types)
def test_service_builder_based_on_custom_config_works(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    max_nodes = 112
    config.defaults.event.max_nodes = max_nodes
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    sut = node.service_builder(service_name).event().create()

    static_config = sut.static_config
    assert static_config.max_nodes == max_nodes



# Copyright (c) 2025 Contributors to the Eclipse Foundation
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Apache Software License 2.0 which is available at
# https://www.apache.org/licenses/LICENSE-2.0, or the MIT license
# which is available at https://opensource.org/licenses/MIT.
#
# SPDX-License-Identifier: Apache-2.0 OR MIT

import ctypes

import iceoryx2 as iox2
import pytest


def test_type_names_are_translated_correctly() -> None:
    assert iox2.get_type_name(ctypes.c_ubyte) == "u8"
    assert iox2.get_type_name(ctypes.c_uint8) == "u8"

    assert iox2.get_type_name(ctypes.c_ushort) == "u16"
    assert iox2.get_type_name(ctypes.c_uint16) == "u16"

    assert iox2.get_type_name(ctypes.c_uint) == "u32"
    assert iox2.get_type_name(ctypes.c_uint32) == "u32"

    assert iox2.get_type_name(ctypes.c_ulong) == "u64"
    assert iox2.get_type_name(ctypes.c_uint64) == "u64"

    assert iox2.get_type_name(ctypes.c_byte) == "i8"
    assert iox2.get_type_name(ctypes.c_int8) == "i8"

    assert iox2.get_type_name(ctypes.c_short) == "i16"
    assert iox2.get_type_name(ctypes.c_int16) == "i16"

    assert iox2.get_type_name(ctypes.c_int) == "i32"
    assert iox2.get_type_name(ctypes.c_int32) == "i32"

    assert iox2.get_type_name(ctypes.c_long) == "i64"
    assert iox2.get_type_name(ctypes.c_int64) == "i64"

    assert iox2.get_type_name(ctypes.c_bool) == "bool"
    assert iox2.get_type_name(ctypes.c_float) == "f32"
    assert iox2.get_type_name(ctypes.c_double) == "f64"



# Copyright (c) 2025 Contributors to the Eclipse Foundation
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Apache Software License 2.0 which is available at
# https://www.apache.org/licenses/LICENSE-2.0, or the MIT license
# which is available at https://opensource.org/licenses/MIT.
#
# SPDX-License-Identifier: Apache-2.0 OR MIT

"""Slice - A class representing a set of contiguous elements of type T."""

import ctypes
from typing import Any, Generic, Type, TypeVar

T = TypeVar("T", bound=ctypes.Structure)


class Slice(Generic[T]):
    """
    A class representing a slice of contiguous elements of type T.

    A Slice provides a view into a contiguous sequence of elements without owning the memory.
    It allows for efficient access and iteration over a portion of a contiguous data structure.

    T -  The type of elements in the slice. Can be const-qualified for read-only slices.
    """

    def __init__(self, data_ptr: int, number_of_elements: int, t: Type[T]) -> None:
        """Initializes a slice with a data_ptr, number_of_elements it contains and the type."""
        self.data_ptr = data_ptr
        self.number_of_elements = number_of_elements
        self.contained_type = t

    def __str__(self) -> str:
        """Returns human-readable string of the contents."""
        return f"Slice {{ data_ptr: {self.data_ptr}, number_of_elements: {self.number_of_elements} }}"

    def __getitem__(self, index: int) -> Any:
        """Acquires a pointer T to the element at the specified index."""
        if not 0 <= index < self.number_of_elements:
            raise IndexError("Slice index out of range")

        typed_ptr = ctypes.cast(self.data_ptr, ctypes.POINTER(self.contained_type))
        return typed_ptr[index]

    def __setitem__(self, index: int, value: T) -> None:
        """Sets the value at the specified index."""
        if not 0 <= index < self.number_of_elements:
            raise IndexError("Slice index out of range")

        typed_ptr = ctypes.cast(self.data_ptr, ctypes.POINTER(self.contained_type))
        typed_ptr[index] = value

    def len(self) -> int:
        """Returns the length / number of elements contained in the `Slice`."""
        return self.number_of_elements

    def as_ptr(self) -> int:
        """Returns a pointer to the first element of the `Slice`."""
        return self.data_ptr



# Copyright (c) 2025 Contributors to the Eclipse Foundation
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Apache Software License 2.0 which is available at
# https://www.apache.org/licenses/LICENSE-2.0, or the MIT license
# which is available at https://opensource.org/licenses/MIT.
#
# SPDX-License-Identifier: Apache-2.0 OR MIT

"""Python language bindings for iceoryx2."""

from ._iceoryx2 import *
from .publish_subscribe_extensions import *
from .request_response_extensions import *
from .slice import Slice



# Copyright (c) 2025 Contributors to the Eclipse Foundation
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Apache Software License 2.0 which is available at
# https://www.apache.org/licenses/LICENSE-2.0, or the MIT license
# which is available at https://opensource.org/licenses/MIT.
#
# SPDX-License-Identifier: Apache-2.0 OR MIT

import ctypes

import iceoryx2 as iox2
import pytest

service_types = [iox2.ServiceType.Ipc, iox2.ServiceType.Local]


class Payload(ctypes.Structure):
    _fields_ = [("data", ctypes.c_ubyte)]


@pytest.mark.parametrize("service_type", service_types)
def test_unable_to_deliver_strategy_can_be_configured(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)
    service_name = iox2.testing.generate_service_name()
    service = (
        node.service_builder(service_name)
        .publish_subscribe(Payload)
        .enable_safe_overflow(False)
        .max_publishers(2)
        .create()
    )

    sut_1 = (
        service.publisher_builder()
        .unable_to_deliver_strategy(iox2.UnableToDeliverStrategy.Block)
        .create()
    )
    sut_2 = (
        service.publisher_builder()
        .unable_to_deliver_strategy(iox2.UnableToDeliverStrategy.DiscardSample)
        .create()
    )

    assert sut_1.unable_to_deliver_strategy == iox2.UnableToDeliverStrategy.Block
    assert (
        sut_2.unable_to_deliver_strategy == iox2.UnableToDeliverStrategy.DiscardSample
    )


@pytest.mark.parametrize("service_type", service_types)
def test_max_loans_can_be_set_up(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)
    service_name = iox2.testing.generate_service_name()
    service = node.service_builder(service_name).publish_subscribe(Payload).create()
    max_loans = 8

    sut = service.publisher_builder().max_loaned_samples(max_loans).create()

    samples = []
    for _ in range(0, max_loans):
        sample = sut.loan_uninit()
        samples.append(sample)

    with pytest.raises(iox2.LoanError):
        sut.loan_uninit()


@pytest.mark.parametrize("service_type", service_types)
def test_deleting_publisher_removes_it_from_the_service(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)
    service_name = iox2.testing.generate_service_name()
    service = (
        node.service_builder(service_name)
        .publish_subscribe(Payload)
        .max_publishers(1)
        .create()
    )

    sut = service.publisher_builder().create()

    with pytest.raises(iox2.PublisherCreateError):
        sut = service.publisher_builder().create()

    sut.delete()

    try:
        sut = service.publisher_builder().create()
    except iox2.PublisherCreateError:
        assert False


@pytest.mark.parametrize("service_type", service_types)
def test_deleting_sample_mut_uninit_releases_it(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)
    service_name = iox2.testing.generate_service_name()
    service = (
        node.service_builder(service_name)
        .publish_subscribe(Payload)
        .max_publishers(1)
        .create()
    )

    sut = service.publisher_builder().max_loaned_samples(1).create()
    sample_uninit = sut.loan_uninit()

    with pytest.raises(iox2.LoanError):
        sample_uninit = sut.loan_uninit()

    sample_uninit.delete()

    try:
        sample_uninit = sut.loan_uninit()
    except iox2.LoanError:
        assert False


@pytest.mark.parametrize("service_type", service_types)
def test_deleting_sample_mut_releases_it(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)
    service_name = iox2.testing.generate_service_name()
    service = (
        node.service_builder(service_name)
        .publish_subscribe(Payload)
        .max_publishers(1)
        .create()
    )

    sut = service.publisher_builder().max_loaned_samples(1).create()
    sample_uninit = sut.loan_uninit()
    sample = sample_uninit.assume_init()

    with pytest.raises(iox2.LoanError):
        sample_uninit = sut.loan_uninit()

    sample.delete()

    try:
        sample_uninit = sut.loan_uninit()
    except iox2.LoanError:
        assert False



# Copyright (c) 2025 Contributors to the Eclipse Foundation
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Apache Software License 2.0 which is available at
# https://www.apache.org/licenses/LICENSE-2.0, or the MIT license
# which is available at https://opensource.org/licenses/MIT.
#
# SPDX-License-Identifier: Apache-2.0 OR MIT

import iceoryx2 as iox2
import pytest


def test_type_name_can_be_constructed() -> None:
    sut = iox2.TypeName.new("hello world")
    assert sut.to_string() == "hello world"


def test_type_name_construction_fails_when_max_len_is_exceeded() -> None:
    sut_value = "x" * (iox2.TypeName.max_len() + 1)
    with pytest.raises(iox2.SemanticStringError):
        iox2.TypeName.new(sut_value)



# Copyright (c) 2025 Contributors to the Eclipse Foundation
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Apache Software License 2.0 which is available at
# https://www.apache.org/licenses/LICENSE-2.0, or the MIT license
# which is available at https://opensource.org/licenses/MIT.
#
# SPDX-License-Identifier: Apache-2.0 OR MIT

import ctypes

import iceoryx2 as iox2
import pytest

service_types = [iox2.ServiceType.Ipc, iox2.ServiceType.Local]


class Payload(ctypes.Structure):
    _fields_ = [("data", ctypes.c_ubyte)]


class LargePayload(ctypes.Structure):
    _fields_ = [("data", ctypes.c_ulonglong)]


@pytest.mark.parametrize("service_type", service_types)
def test_send_and_receive_request_with_memmove_works(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)
    number_of_requests = 5

    service_name = iox2.testing.generate_service_name()
    service = (
        node.service_builder(service_name)
        .request_response(Payload, Payload)
        .max_active_requests_per_client(number_of_requests)
        .enable_fire_and_forget_requests(True)
        .create()
    )

    client = service.client_builder().create()
    server = service.server_builder().create()

    for i in range(0, number_of_requests):
        request_payload = Payload(data=19 + i)
        request_uninit = client.loan_uninit()
        ctypes.memmove(
            request_uninit.payload_ptr,
            ctypes.byref(request_payload),
            ctypes.sizeof(Payload),
        )
        request = request_uninit.assume_init()
        request.send()

    for i in range(0, number_of_requests):
        assert server.has_requests
        request = server.receive()
        assert request is not None
        request_payload = Payload(data=0)
        ctypes.memmove(
            ctypes.byref(request_payload),
            request.payload_ptr,
            ctypes.sizeof(Payload),
        )
        assert request_payload.data == 19 + i

    assert not server.has_requests


@pytest.mark.parametrize("service_type", service_types)
def test_send_and_receive_responses_with_memmove_works(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)
    number_of_responses = 5

    service_name = iox2.testing.generate_service_name()
    service = (
        node.service_builder(service_name)
        .request_response(Payload, Payload)
        .max_response_buffer_size(number_of_responses)
        .enable_fire_and_forget_requests(True)
        .create()
    )

    client = service.client_builder().create()
    server = service.server_builder().create()
    pending_response = client.send_copy(Payload(data=12))

    active_request = server.receive()

    for i in range(0, number_of_responses):
        response_payload = Payload(data=3 + 2 * i)
        response_uninit = active_request.loan_uninit()
        ctypes.memmove(
            response_uninit.payload_ptr,
            ctypes.byref(response_payload),
            ctypes.sizeof(Payload),
        )
        response = response_uninit.assume_init()
        response.send()

    for i in range(0, number_of_responses):
        assert pending_response.has_response
        response = pending_response.receive()
        payload = Payload(data=0)
        ctypes.memmove(
            ctypes.byref(payload), response.payload_ptr, ctypes.sizeof(Payload)
        )
        assert payload.data == 3 + 2 * i

    assert not pending_response.has_response


@pytest.mark.parametrize("service_type", service_types)
def test_send_and_receive_request_with_sendcopy_works(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    service = (
        node.service_builder(service_name).request_response(Payload, Payload).create()
    )

    client = service.client_builder().create()
    server = service.server_builder().create()

    pending_response = client.send_copy(Payload(data=87))
    active_request = server.receive()

    assert active_request.payload().contents.data == 87
    active_request.send_copy(Payload(data=33))

    response = pending_response.receive()
    assert response.payload().contents.data == 33


@pytest.mark.parametrize("service_type", service_types)
def test_send_with_request_user_header_works(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    service = (
        node.service_builder(service_name)
        .request_response(Payload, Payload)
        .request_header(ctypes.c_uint64)
        .create()
    )

    client = service.client_builder().create()
    server = service.server_builder().create()

    request_uninit = client.loan_uninit()
    ctypes.memmove(request_uninit.user_header_ptr, ctypes.byref(ctypes.c_uint64(89)), 8)
    assert request_uninit.user_header().contents.value == 89

    request = request_uninit.assume_init()
    assert request.user_header().contents.value == 89
    _pending_response = request.send()

    active_request = server.receive()
    assert active_request.user_header().contents.value == 89


@pytest.mark.parametrize("service_type", service_types)
def test_send_with_response_user_header_works(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    service = (
        node.service_builder(service_name)
        .request_response(Payload, Payload)
        .response_header(ctypes.c_uint64)
        .create()
    )

    client = service.client_builder().create()
    server = service.server_builder().create()

    request_uninit = client.loan_uninit()
    request = request_uninit.assume_init()
    pending_response = request.send()

    active_request = server.receive()
    response_uninit = active_request.loan_uninit()
    ctypes.memmove(
        response_uninit.user_header_ptr, ctypes.byref(ctypes.c_uint64(44)), 8
    )
    assert response_uninit.user_header().contents.value == 44
    response = response_uninit.assume_init()
    assert response.user_header().contents.value == 44
    response.send()

    response = pending_response.receive()
    assert response.user_header().contents.value == 44


@pytest.mark.parametrize("service_type", service_types)
def test_send_with_request_system_header_works(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    service = (
        node.service_builder(service_name).request_response(Payload, Payload).create()
    )

    client = service.client_builder().create()
    server = service.server_builder().create()

    request_uninit = client.loan_uninit()

    assert request_uninit.header.client_id == client.id
    assert request_uninit.header.number_of_elements == 1

    request = request_uninit.assume_init()

    assert request.header.client_id == client.id
    assert request.header.number_of_elements == 1

    _pending_response = request.send()

    active_request = server.receive()

    assert active_request.header.client_id == client.id
    assert active_request.header.number_of_elements == 1


@pytest.mark.parametrize("service_type", service_types)
def test_send_with_response_system_header_works(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    service = (
        node.service_builder(service_name).request_response(Payload, Payload).create()
    )

    client = service.client_builder().create()
    server = service.server_builder().create()

    request_uninit = client.loan_uninit()
    request = request_uninit.assume_init()
    pending_response = request.send()

    active_request = server.receive()
    response_uninit = active_request.loan_uninit()

    assert response_uninit.header.server_id == server.id
    assert response_uninit.header.number_of_elements == 1

    response = response_uninit.assume_init()

    assert response.header.server_id == server.id
    assert response.header.number_of_elements == 1

    response.send()

    response = pending_response.receive()

    assert response.header.server_id == server.id
    assert response.header.number_of_elements == 1


@pytest.mark.parametrize("service_type", service_types)
def test_client_reallocation_fails_when_allocation_strategy_is_static(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    service = (
        node.service_builder(service_name)
        .request_response(iox2.Slice[ctypes.c_uint8], iox2.Slice[ctypes.c_uint8])
        .create()
    )

    client = (
        service.client_builder()
        .initial_max_slice_len(8)
        .allocation_strategy(iox2.AllocationStrategy.Static)
        .create()
    )

    try:
        client.loan_slice_uninit(8)
    except iox2.LoanError:
        assert False

    with pytest.raises(iox2.LoanError):
        client.loan_slice_uninit(9)


@pytest.mark.parametrize("service_type", service_types)
def test_client_reallocation_works_when_allocation_strategy_is_not_static(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    service = (
        node.service_builder(service_name)
        .request_response(iox2.Slice[ctypes.c_uint8], iox2.Slice[ctypes.c_uint8])
        .create()
    )

    client = (
        service.client_builder()
        .initial_max_slice_len(8)
        .allocation_strategy(iox2.AllocationStrategy.PowerOfTwo)
        .create()
    )

    try:
        client.loan_slice_uninit(12)
    except iox2.LoanError:
        assert False


@pytest.mark.parametrize("service_type", service_types)
def test_server_reallocation_fails_when_allocation_strategy_is_static(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    service = (
        node.service_builder(service_name)
        .request_response(Payload, iox2.Slice[ctypes.c_uint8])
        .create()
    )

    client = service.client_builder().create()
    server = (
        service.server_builder()
        .initial_max_slice_len(8)
        .allocation_strategy(iox2.AllocationStrategy.Static)
        .create()
    )

    _pending_response = client.send_copy(Payload(data=1))
    active_request = server.receive()

    with pytest.raises(iox2.LoanError):
        active_request.loan_slice_uninit(9)

    try:
        active_request.loan_slice_uninit(8)
    except iox2.LoanError:
        assert False


@pytest.mark.parametrize("service_type", service_types)
def test_server_reallocation_works_when_allocation_strategy_is_not_static(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    service = (
        node.service_builder(service_name)
        .request_response(Payload, iox2.Slice[ctypes.c_uint8])
        .create()
    )

    client = service.client_builder().create()
    server = (
        service.server_builder()
        .initial_max_slice_len(8)
        .allocation_strategy(iox2.AllocationStrategy.PowerOfTwo)
        .create()
    )

    _pending_response = client.send_copy(Payload(data=1))
    active_request = server.receive()

    try:
        active_request.loan_slice_uninit(12)
    except iox2.LoanError:
        assert False


@pytest.mark.parametrize("service_type", service_types)
def test_slice_type_forbids_use_of_non_slice_api(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    service = (
        node.service_builder(service_name)
        .request_response(iox2.Slice[ctypes.c_uint8], iox2.Slice[ctypes.c_uint8])
        .create()
    )

    server = service.server_builder().create()
    client = service.client_builder().create()

    with pytest.raises(AssertionError):
        client.loan_uninit()

    request_uninit = client.loan_slice_uninit(1)
    request = request_uninit.assume_init()
    request.send()

    active_request = server.receive()

    with pytest.raises(AssertionError):
        active_request.loan_uninit()


@pytest.mark.parametrize("service_type", service_types)
def test_non_slice_type_forbids_use_of_slice_api(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    service = (
        node.service_builder(service_name).request_response(Payload, Payload).create()
    )

    server = service.server_builder().create()
    client = service.client_builder().create()

    with pytest.raises(AssertionError):
        client.loan_slice_uninit(1)

    request_uninit = client.loan_uninit()
    request = request_uninit.assume_init()
    request.send()

    active_request = server.receive()

    with pytest.raises(AssertionError):
        active_request.loan_slice_uninit(1)


@pytest.mark.parametrize("service_type", service_types)
def test_client_can_request_graceful_disconnect(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    service = (
        node.service_builder(service_name).request_response(Payload, Payload).create()
    )

    client = service.client_builder().create()
    server = service.server_builder().create()

    pending_response = client.send_copy(Payload(data=0))
    active_request = server.receive()

    assert pending_response.is_connected is True
    assert active_request.is_connected is True
    assert active_request.has_disconnect_hint is False

    pending_response.set_disconnect_hint()

    assert pending_response.is_connected is True
    assert active_request.is_connected is True
    assert active_request.has_disconnect_hint is True

    pending_response.delete()

    assert active_request.is_connected is False
    assert active_request.has_disconnect_hint is False



# Copyright (c) 2025 Contributors to the Eclipse Foundation
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Apache Software License 2.0 which is available at
# https://www.apache.org/licenses/LICENSE-2.0, or the MIT license
# which is available at https://opensource.org/licenses/MIT.
#
# SPDX-License-Identifier: Apache-2.0 OR MIT

import iceoryx2 as iox2
import pytest


def test_global_root_path_can_be_set() -> None:
    sut = iox2.config.default()
    path = iox2.Path.new("/some/path")
    sut.global_cfg.root_path = path
    assert sut.global_cfg.root_path == path


def test_global_prefix_can_be_set() -> None:
    sut = iox2.config.default()
    path = iox2.FileName.new("prefix_")
    sut.global_cfg.prefix = path
    assert sut.global_cfg.prefix == path


def test_global_service_directory_can_be_set() -> None:
    sut = iox2.config.default()
    path = iox2.Path.new("/path/to/funky")
    sut.global_cfg.service.directory = path
    assert sut.global_cfg.service.directory == path


def test_global_service_data_segment_suffix_can_be_set() -> None:
    sut = iox2.config.default()
    path = iox2.FileName.new(".let_suf_the_fix")
    sut.global_cfg.service.data_segment_suffix = path
    assert sut.global_cfg.service.data_segment_suffix == path


def test_global_service_static_config_storage_suffix_can_be_set() -> None:
    sut = iox2.config.default()
    path = iox2.FileName.new(".its_a_fix_for_the_suf")
    sut.global_cfg.service.static_config_storage_suffix = path
    assert sut.global_cfg.service.static_config_storage_suffix == path


def test_global_service_dynamic_config_storage_suffix_can_be_set() -> None:
    sut = iox2.config.default()
    path = iox2.FileName.new(".flufn_fluff")
    sut.global_cfg.service.dynamic_config_storage_suffix = path
    assert sut.global_cfg.service.dynamic_config_storage_suffix == path


def test_global_service_creation_timeout_can_be_set() -> None:
    sut = iox2.config.default()
    path = iox2.Duration.from_millis(123)
    sut.global_cfg.service.creation_timeout = path
    assert sut.global_cfg.service.creation_timeout == path


def test_global_service_connection_suffix_can_be_set() -> None:
    sut = iox2.config.default()
    path = iox2.FileName.new(".fuldaba")
    sut.global_cfg.service.connection_suffix = path
    assert sut.global_cfg.service.connection_suffix == path


def test_global_service_event_connection_suffix_can_be_set() -> None:
    sut = iox2.config.default()
    path = iox2.FileName.new(".blarb")
    sut.global_cfg.service.event_connection_suffix = path
    assert sut.global_cfg.service.event_connection_suffix == path


def test_global_node_directory_can_be_set() -> None:
    sut = iox2.config.default()
    value = iox2.Path.new("/dir/to/somewher")
    sut.global_cfg.node.directory = value
    assert sut.global_cfg.node.directory == value


def test_global_node_monitor_suffix_can_be_set() -> None:
    sut = iox2.config.default()
    value = iox2.FileName.new(".suf_le_fix_le")
    sut.global_cfg.node.monitor_suffix = value
    assert sut.global_cfg.node.monitor_suffix == value


def test_global_node_static_config_suffix_can_be_set() -> None:
    sut = iox2.config.default()
    value = iox2.FileName.new(".wuff")
    sut.global_cfg.node.static_config_suffix = value
    assert sut.global_cfg.node.static_config_suffix == value


def test_global_node_service_tag_suffix_can_be_set() -> None:
    sut = iox2.config.default()
    value = iox2.FileName.new(".grogg")
    sut.global_cfg.node.service_tag_suffix = value
    assert sut.global_cfg.node.service_tag_suffix == value


def test_global_node_cleanup_dead_nodes_on_destruction_can_be_set() -> None:
    sut = iox2.config.default()
    sut.global_cfg.node.cleanup_dead_nodes_on_destruction = True
    assert sut.global_cfg.node.cleanup_dead_nodes_on_destruction
    sut.global_cfg.node.cleanup_dead_nodes_on_destruction = False
    assert not sut.global_cfg.node.cleanup_dead_nodes_on_destruction


def test_global_node_cleanup_dead_nodes_on_creation_can_be_set() -> None:
    sut = iox2.config.default()
    sut.global_cfg.node.cleanup_dead_nodes_on_creation = True
    assert sut.global_cfg.node.cleanup_dead_nodes_on_creation
    sut.global_cfg.node.cleanup_dead_nodes_on_creation = False
    assert not sut.global_cfg.node.cleanup_dead_nodes_on_creation


def test_defaults_request_response_safe_overflow_for_requests_can_be_set() -> None:
    sut = iox2.config.default()
    sut.defaults.request_response.enable_safe_overflow_for_requests = False
    assert not sut.defaults.request_response.enable_safe_overflow_for_requests
    sut.defaults.request_response.enable_safe_overflow_for_requests = True
    assert sut.defaults.request_response.enable_safe_overflow_for_requests


def test_defaults_request_response_safe_overflow_for_responses_can_be_set() -> None:
    sut = iox2.config.default()
    sut.defaults.request_response.enable_safe_overflow_for_responses = False
    assert not sut.defaults.request_response.enable_safe_overflow_for_responses
    sut.defaults.request_response.enable_safe_overflow_for_responses = True
    assert sut.defaults.request_response.enable_safe_overflow_for_responses


def test_defaults_request_response_max_active_requests_per_client_can_be_set() -> None:
    sut = iox2.config.default()
    value = 912
    sut.defaults.request_response.max_active_requests_per_client = value
    assert sut.defaults.request_response.max_active_requests_per_client == value


def test_defaults_request_response_max_response_buffer_size_can_be_set() -> None:
    sut = iox2.config.default()
    value = 281
    sut.defaults.request_response.max_response_buffer_size = value
    assert sut.defaults.request_response.max_response_buffer_size == value


def test_defaults_request_response_max_servers_can_be_set() -> None:
    sut = iox2.config.default()
    value = 91
    sut.defaults.request_response.max_servers = value
    assert sut.defaults.request_response.max_servers == value


def test_defaults_request_response_max_clients_can_be_set() -> None:
    sut = iox2.config.default()
    value = 661
    sut.defaults.request_response.max_clients = value
    assert sut.defaults.request_response.max_clients == value


def test_defaults_request_response_max_nodes_can_be_set() -> None:
    sut = iox2.config.default()
    value = 45
    sut.defaults.request_response.max_nodes = value
    assert sut.defaults.request_response.max_nodes == value


def test_defaults_request_response_max_borrowed_responses_per_pending_response_can_be_set() -> (
    None
):
    sut = iox2.config.default()
    value = 23
    sut.defaults.request_response.max_borrowed_responses_per_pending_response = value
    assert (
        sut.defaults.request_response.max_borrowed_responses_per_pending_response
        == value
    )


def test_defaults_request_response_max_loaned_requests_can_be_set() -> None:
    sut = iox2.config.default()
    value = 41
    sut.defaults.request_response.max_loaned_requests = value
    assert sut.defaults.request_response.max_loaned_requests == value


def test_defaults_request_response_server_max_loaned_responses_per_request_can_be_set() -> (
    None
):
    sut = iox2.config.default()
    value = 56
    sut.defaults.request_response.server_max_loaned_responses_per_request = value
    assert (
        sut.defaults.request_response.server_max_loaned_responses_per_request == value
    )


def test_defaults_request_response_client_unable_to_deliver_strategy_can_be_set() -> (
    None
):
    sut = iox2.config.default()
    sut.defaults.request_response.client_unable_to_deliver_strategy = (
        iox2.UnableToDeliverStrategy.Block
    )
    assert (
        sut.defaults.request_response.client_unable_to_deliver_strategy
        == iox2.UnableToDeliverStrategy.Block
    )
    sut.defaults.request_response.client_unable_to_deliver_strategy = (
        iox2.UnableToDeliverStrategy.DiscardSample
    )
    assert (
        sut.defaults.request_response.client_unable_to_deliver_strategy
        == iox2.UnableToDeliverStrategy.DiscardSample
    )


def test_defaults_request_response_server_unable_to_deliver_strategy_can_be_set() -> (
    None
):
    sut = iox2.config.default()
    sut.defaults.request_response.server_unable_to_deliver_strategy = (
        iox2.UnableToDeliverStrategy.Block
    )
    assert (
        sut.defaults.request_response.server_unable_to_deliver_strategy
        == iox2.UnableToDeliverStrategy.Block
    )
    sut.defaults.request_response.server_unable_to_deliver_strategy = (
        iox2.UnableToDeliverStrategy.DiscardSample
    )
    assert (
        sut.defaults.request_response.server_unable_to_deliver_strategy
        == iox2.UnableToDeliverStrategy.DiscardSample
    )


def test_defaults_request_response_client_expired_connection_buffer_can_be_set() -> (
    None
):
    sut = iox2.config.default()
    value = 78
    sut.defaults.request_response.client_expired_connection_buffer = value
    assert sut.defaults.request_response.client_expired_connection_buffer == value


def test_defaults_request_response_server_expired_connection_buffer_can_be_set() -> (
    None
):
    sut = iox2.config.default()
    value = 213
    sut.defaults.request_response.server_expired_connection_buffer = value
    assert sut.defaults.request_response.server_expired_connection_buffer == value


def test_defaults_request_response_enable_fire_and_forget_requests_can_be_set() -> None:
    sut = iox2.config.default()
    sut.defaults.request_response.enable_fire_and_forget_requests = True
    assert sut.defaults.request_response.enable_fire_and_forget_requests
    sut.defaults.request_response.enable_fire_and_forget_requests = False
    assert not sut.defaults.request_response.enable_fire_and_forget_requests


def test_defaults_event_max_listeners_can_be_set() -> None:
    sut = iox2.config.default()
    value = 8891
    sut.defaults.event.max_listeners = value
    assert sut.defaults.event.max_listeners == value


def test_defaults_event_max_notifiers_can_be_set() -> None:
    sut = iox2.config.default()
    value = 3321
    sut.defaults.event.max_notifiers = value
    assert sut.defaults.event.max_notifiers == value


def test_defaults_event_max_nodes_can_be_set() -> None:
    sut = iox2.config.default()
    value = 1121
    sut.defaults.event.max_nodes = value
    assert sut.defaults.event.max_nodes == value


def test_defaults_event_event_id_max_value_can_be_set() -> None:
    sut = iox2.config.default()
    value = 18
    sut.defaults.event.event_id_max_value = value
    assert sut.defaults.event.event_id_max_value == value


def test_defaults_event_deadline_can_be_set_can_be_set() -> None:
    sut = iox2.config.default()
    value = iox2.Duration.from_secs(2)
    sut.defaults.event.deadline = value
    assert sut.defaults.event.deadline == value


def test_defaults_event_notifier_created_event_can_be_set() -> None:
    sut = iox2.config.default()
    value = 941
    sut.defaults.event.notifier_created_event = value
    assert sut.defaults.event.notifier_created_event == value
    assert sut.defaults.event.has_notifier_created_event
    sut.defaults.event.disable_notifier_created_event()
    assert not sut.defaults.event.has_notifier_created_event


def test_defaults_event_notifier_dropped_event_can_be_set() -> None:
    sut = iox2.config.default()
    value = 9411
    sut.defaults.event.notifier_dropped_event = value
    assert sut.defaults.event.notifier_dropped_event == value
    assert sut.defaults.event.has_notifier_dropped_event
    sut.defaults.event.disable_notifier_dropped_event()
    assert not sut.defaults.event.has_notifier_dropped_event


def test_defaults_event_notifier_dead_event_can_be_set() -> None:
    sut = iox2.config.default()
    value = 9411
    sut.defaults.event.notifier_dead_event = value
    assert sut.defaults.event.notifier_dead_event == value
    assert sut.defaults.event.has_notifier_dead_event
    sut.defaults.event.disable_notifier_dead_event()
    assert not sut.defaults.event.has_notifier_dead_event


def test_defaults_publish_subscribe_max_subscribers_can_be_set() -> None:
    sut = iox2.config.default()
    value = 1818
    sut.defaults.publish_subscribe.max_subscribers = value
    assert sut.defaults.publish_subscribe.max_subscribers == value


def test_defaults_publish_subscribe_max_publishers_can_be_set() -> None:
    sut = iox2.config.default()
    value = 8181
    sut.defaults.publish_subscribe.max_publishers = value
    assert sut.defaults.publish_subscribe.max_publishers == value


def test_defaults_publish_subscribe_max_nodes_can_be_set() -> None:
    sut = iox2.config.default()
    value = 5454
    sut.defaults.publish_subscribe.max_nodes = value
    assert sut.defaults.publish_subscribe.max_nodes == value


def test_defaults_publish_subscribe_subscriber_max_buffer_size_can_be_set() -> None:
    sut = iox2.config.default()
    value = 9918
    sut.defaults.publish_subscribe.subscriber_max_buffer_size = value
    assert sut.defaults.publish_subscribe.subscriber_max_buffer_size == value


def test_defaults_publish_subscribe_subscriber_max_borrowed_samples_can_be_set() -> (
    None
):
    sut = iox2.config.default()
    value = 23786
    sut.defaults.publish_subscribe.subscriber_max_borrowed_samples = value
    assert sut.defaults.publish_subscribe.subscriber_max_borrowed_samples == value


def test_defaults_publish_subscribe_publisher_max_loaned_samples_can_be_set() -> None:
    sut = iox2.config.default()
    value = 182
    sut.defaults.publish_subscribe.publisher_max_loaned_samples = value
    assert sut.defaults.publish_subscribe.publisher_max_loaned_samples == value


def test_defaults_publish_subscribe_publisher_history_size_can_be_set() -> None:
    sut = iox2.config.default()
    value = 18221
    sut.defaults.publish_subscribe.publisher_history_size = value
    assert sut.defaults.publish_subscribe.publisher_history_size == value


def test_defaults_publish_subscribe_enable_safe_overflow_can_be_set() -> None:
    sut = iox2.config.default()
    sut.defaults.publish_subscribe.enable_safe_overflow = True
    assert sut.defaults.publish_subscribe.enable_safe_overflow
    sut.defaults.publish_subscribe.enable_safe_overflow = False
    assert not sut.defaults.publish_subscribe.enable_safe_overflow


def test_defaults_publish_subscribe_unable_to_deliver_strategy_can_be_set() -> None:
    sut = iox2.config.default()
    sut.defaults.publish_subscribe.unable_to_deliver_strategy = (
        iox2.UnableToDeliverStrategy.Block
    )
    assert (
        sut.defaults.publish_subscribe.unable_to_deliver_strategy
        == iox2.UnableToDeliverStrategy.Block
    )
    sut.defaults.publish_subscribe.unable_to_deliver_strategy = (
        iox2.UnableToDeliverStrategy.DiscardSample
    )
    assert (
        sut.defaults.publish_subscribe.unable_to_deliver_strategy
        == iox2.UnableToDeliverStrategy.DiscardSample
    )


def test_defaults_publish_subscribe_subscriber_expired_connection_buffer_can_be_set() -> (
    None
):
    sut = iox2.config.default()
    value = 56273
    sut.defaults.publish_subscribe.subscriber_expired_connection_buffer = value
    assert sut.defaults.publish_subscribe.subscriber_expired_connection_buffer == value



# Copyright (c) 2025 Contributors to the Eclipse Foundation
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Apache Software License 2.0 which is available at
# https://www.apache.org/licenses/LICENSE-2.0, or the MIT license
# which is available at https://opensource.org/licenses/MIT.
#
# SPDX-License-Identifier: Apache-2.0 OR MIT

import ctypes

import iceoryx2 as iox2
import pytest

service_types = [iox2.ServiceType.Ipc, iox2.ServiceType.Local]


class Payload(ctypes.Structure):
    _fields_ = [("data", ctypes.c_ubyte)]


@pytest.mark.parametrize("service_type", service_types)
def test_unable_to_deliver_strategy_can_be_configured(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)
    service_name = iox2.testing.generate_service_name()
    service = (
        node.service_builder(service_name)
        .request_response(Payload, Payload)
        .enable_safe_overflow_for_requests(False)
        .max_clients(2)
        .create()
    )

    sut_1 = (
        service.client_builder()
        .unable_to_deliver_strategy(iox2.UnableToDeliverStrategy.Block)
        .create()
    )
    sut_2 = (
        service.client_builder()
        .unable_to_deliver_strategy(iox2.UnableToDeliverStrategy.DiscardSample)
        .create()
    )

    assert sut_1.unable_to_deliver_strategy == iox2.UnableToDeliverStrategy.Block
    assert (
        sut_2.unable_to_deliver_strategy == iox2.UnableToDeliverStrategy.DiscardSample
    )


@pytest.mark.parametrize("service_type", service_types)
def test_deleting_client_removes_it_from_the_service(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)
    service_name = iox2.testing.generate_service_name()
    service = (
        node.service_builder(service_name)
        .request_response(Payload, Payload)
        .max_clients(1)
        .create()
    )

    sut = service.client_builder().create()

    with pytest.raises(iox2.ClientCreateError):
        sut = service.client_builder().create()

    sut.delete()

    try:
        sut = service.client_builder().create()
    except iox2.ClientCreateError:
        assert False


@pytest.mark.parametrize("service_type", service_types)
def test_deleting_request_mut_uninit_removes_it_from_the_service(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)
    service_name = iox2.testing.generate_service_name()
    service = (
        node.service_builder(service_name)
        .request_response(Payload, Payload)
        .max_loaned_requests(1)
        .create()
    )

    sut = service.client_builder().create()
    request_uninit = sut.loan_uninit()

    with pytest.raises(iox2.LoanError):
        request_uninit = sut.loan_uninit()

    request_uninit.delete()

    try:
        request_uninit = sut.loan_uninit()
    except iox2.LoanError:
        assert False


@pytest.mark.parametrize("service_type", service_types)
def test_deleting_request_mut_removes_it_from_the_service(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)
    service_name = iox2.testing.generate_service_name()
    service = (
        node.service_builder(service_name)
        .request_response(Payload, Payload)
        .max_loaned_requests(1)
        .create()
    )

    sut = service.client_builder().create()
    request_uninit = sut.loan_uninit()
    request = request_uninit.assume_init()

    with pytest.raises(iox2.LoanError):
        request_uninit = sut.loan_uninit()

    request.delete()

    try:
        request_uninit = sut.loan_uninit()
    except iox2.LoanError:
        assert False



# Copyright (c) 2025 Contributors to the Eclipse Foundation
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Apache Software License 2.0 which is available at
# https://www.apache.org/licenses/LICENSE-2.0, or the MIT license
# which is available at https://opensource.org/licenses/MIT.
#
# SPDX-License-Identifier: Apache-2.0 OR MIT

import iceoryx2 as iox2
import pytest

service_types = [iox2.ServiceType.Ipc, iox2.ServiceType.Local]


@pytest.mark.parametrize("service_type", service_types)
def test_notifier_use_default_event_id(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)
    event_id = iox2.EventId.new(45)

    service_name = iox2.testing.generate_service_name()
    service = node.service_builder(service_name).event().create()

    notifier = service.notifier_builder().default_event_id(event_id).create()
    listener = service.listener_builder().create()

    notifier.notify()
    assert listener.try_wait_one() == event_id


@pytest.mark.parametrize("service_type", service_types)
def test_notification_with_custom_event_id_works(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)
    event_id = iox2.EventId.new(41)

    service_name = iox2.testing.generate_service_name()
    service = node.service_builder(service_name).event().create()

    notifier = service.notifier_builder().create()
    listener = service.listener_builder().create()

    notifier.notify_with_custom_event_id(event_id)
    assert listener.try_wait_one() == event_id


@pytest.mark.parametrize("service_type", service_types)
def test_deadline_can_be_acquired_via_ports(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)
    deadline = iox2.Duration.from_secs(123)

    service_name = iox2.testing.generate_service_name()
    service = node.service_builder(service_name).event().deadline(deadline).create()

    notifier = service.notifier_builder().create()
    listener = service.listener_builder().create()

    assert listener.deadline == deadline
    assert notifier.deadline == deadline


@pytest.mark.parametrize("service_type", service_types)
def test_listener_try_wait_one_works(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)
    event_id_1 = iox2.EventId.new(12)
    event_id_2 = iox2.EventId.new(21)

    service_name = iox2.testing.generate_service_name()
    service = node.service_builder(service_name).event().create()

    notifier = service.notifier_builder().create()
    listener = service.listener_builder().create()

    notifier.notify_with_custom_event_id(event_id_1)
    notifier.notify_with_custom_event_id(event_id_2)

    assert listener.try_wait_one() == event_id_1
    assert listener.try_wait_one() == event_id_2


@pytest.mark.parametrize("service_type", service_types)
def test_listener_timed_wait_one_works(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)
    timeout = iox2.Duration.from_secs(1)
    event_id_1 = iox2.EventId.new(13)
    event_id_2 = iox2.EventId.new(31)

    service_name = iox2.testing.generate_service_name()
    service = node.service_builder(service_name).event().create()

    notifier = service.notifier_builder().create()
    listener = service.listener_builder().create()

    notifier.notify_with_custom_event_id(event_id_1)
    notifier.notify_with_custom_event_id(event_id_2)

    assert listener.timed_wait_one(timeout) == event_id_1
    assert listener.timed_wait_one(timeout) == event_id_2


@pytest.mark.parametrize("service_type", service_types)
def test_listener_blocking_wait_one_works(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)
    event_id_1 = iox2.EventId.new(14)
    event_id_2 = iox2.EventId.new(41)

    service_name = iox2.testing.generate_service_name()
    service = node.service_builder(service_name).event().create()

    notifier = service.notifier_builder().create()
    listener = service.listener_builder().create()

    notifier.notify_with_custom_event_id(event_id_1)
    notifier.notify_with_custom_event_id(event_id_2)

    assert listener.blocking_wait_one() == event_id_1
    assert listener.blocking_wait_one() == event_id_2


@pytest.mark.parametrize("service_type", service_types)
def test_listener_try_wait_all_works(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)
    event_id_1 = iox2.EventId.new(15)
    event_id_2 = iox2.EventId.new(51)

    service_name = iox2.testing.generate_service_name()
    service = node.service_builder(service_name).event().create()

    notifier = service.notifier_builder().create()
    listener = service.listener_builder().create()

    notifier.notify_with_custom_event_id(event_id_1)
    notifier.notify_with_custom_event_id(event_id_2)

    events = listener.try_wait_all()

    assert len(events) == 2
    assert events[0] == event_id_1
    assert events[1] == event_id_2


@pytest.mark.parametrize("service_type", service_types)
def test_listener_timed_wait_all_works(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)
    event_id_1 = iox2.EventId.new(16)
    event_id_2 = iox2.EventId.new(61)
    timeout = iox2.Duration.from_secs(2)

    service_name = iox2.testing.generate_service_name()
    service = node.service_builder(service_name).event().create()

    notifier = service.notifier_builder().create()
    listener = service.listener_builder().create()

    notifier.notify_with_custom_event_id(event_id_1)
    notifier.notify_with_custom_event_id(event_id_2)

    events = listener.timed_wait_all(timeout)

    assert len(events) == 2
    assert events[0] == event_id_1
    assert events[1] == event_id_2


@pytest.mark.parametrize("service_type", service_types)
def test_listener_blocking_wait_all_works(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)
    event_id_1 = iox2.EventId.new(17)
    event_id_2 = iox2.EventId.new(71)

    service_name = iox2.testing.generate_service_name()
    service = node.service_builder(service_name).event().create()

    notifier = service.notifier_builder().create()
    listener = service.listener_builder().create()

    notifier.notify_with_custom_event_id(event_id_1)
    notifier.notify_with_custom_event_id(event_id_2)

    events = listener.blocking_wait_all()

    assert len(events) == 2
    assert events[0] == event_id_1
    assert events[1] == event_id_2



# Copyright (c) 2025 Contributors to the Eclipse Foundation
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Apache Software License 2.0 which is available at
# https://www.apache.org/licenses/LICENSE-2.0, or the MIT license
# which is available at https://opensource.org/licenses/MIT.
#
# SPDX-License-Identifier: Apache-2.0 OR MIT

"""Strong type safe extensions for the publish-subscribe messaging pattern."""

import ctypes
from typing import Any, Type, TypeVar, get_args, get_origin

from ._iceoryx2 import *
from .slice import Slice
from .type_name import get_type_name

T = TypeVar("T", bound=ctypes.Structure)


def payload(self: Any) -> Any:
    """Returns a `ctypes.POINTER` to the payload."""
    assert self.__payload_type_details is not None
    if get_origin(self.__payload_type_details) is Slice:
        (contained_type,) = get_args(self.__payload_type_details)
        return Slice(self.payload_ptr, self.__slice_len, contained_type)

    return ctypes.cast(self.payload_ptr, ctypes.POINTER(self.__payload_type_details))


def user_header(self: Any) -> Any:
    """Returns a `ctypes.POINTER` to the user header."""
    assert self.__user_header_type_details is not None
    return ctypes.cast(
        self.user_header_ptr, ctypes.POINTER(self.__user_header_type_details)
    )


def publish_subscribe(
    self: ServiceBuilder, t: Type[T]
) -> ServiceBuilderPublishSubscribe:
    """Returns the `ServiceBuilderPublishSusbcribe` to create a new publish-subscribe service. The payload ctype must be provided as argument."""
    type_name = t.__name__
    type_size = 0
    type_align = 0
    type_variant = TypeVariant.FixedSize

    if get_origin(t) is Slice:
        (contained_type,) = get_args(t)
        type_name = get_type_name(contained_type)
        type_variant = TypeVariant.Dynamic
        type_size = ctypes.sizeof(contained_type)
        type_align = ctypes.alignment(contained_type)
    else:
        type_name = get_type_name(t)
        type_size = ctypes.sizeof(t)
        type_align = ctypes.alignment(t)
        type_variant = TypeVariant.FixedSize

    result = self.__publish_subscribe()
    result.__set_payload_type(t)

    return result.__payload_type_details(
        TypeDetail.new()
        .type_variant(type_variant)
        .type_name(TypeName.new(type_name))
        .size(type_size)
        .alignment(type_align)
    ).__user_header_type_details(
        TypeDetail.new()
        .type_variant(TypeVariant.FixedSize)
        .type_name(TypeName.new("()"))
        .size(0)
        .alignment(1)
    )


def set_user_header(
    self: ServiceBuilderPublishSubscribe, t: Type[T]
) -> ServiceBuilderPublishSubscribe:
    """Sets the user header type for the service."""
    type_name = get_type_name(t)
    result = self.__user_header_type_details(
        TypeDetail.new()
        .type_variant(TypeVariant.FixedSize)
        .type_name(TypeName.new(type_name))
        .size(ctypes.sizeof(t))
        .alignment(ctypes.alignment(t))
    )
    result.__set_user_header_type(t)
    return result


def send_copy(self: Publisher, t: Type[T]) -> Any:
    """Sends a copy of the provided type."""
    assert self.__payload_type_details is not None
    sample_uninit = self.loan_uninit()

    assert ctypes.sizeof(t) == ctypes.sizeof(sample_uninit.__payload_type_details)
    assert ctypes.alignment(t) == ctypes.alignment(sample_uninit.__payload_type_details)

    ctypes.memmove(sample_uninit.payload_ptr, ctypes.byref(t), ctypes.sizeof(t))
    sample = sample_uninit.assume_init()
    return sample.send()


def write_payload(self: SampleMutUninit, t: Type[T]) -> SampleMut:
    """Writes the provided payload into the sample."""
    assert self.__payload_type_details is not None
    assert ctypes.sizeof(t) == ctypes.sizeof(self.__payload_type_details)
    assert ctypes.alignment(t) == ctypes.alignment(self.__payload_type_details)

    ctypes.memmove(self.payload_ptr, ctypes.byref(t), ctypes.sizeof(t))
    return self.assume_init()


def loan_uninit(self: Publisher) -> SampleMutUninit:
    """
    Loans/allocates a `SampleMutUninit` from the underlying data segment of the `Publisher`.

    The user has to initialize the payload before it can be sent. On failure it returns
    `LoanError` describing the failure.
    """
    assert not get_origin(self.__payload_type_details) is Slice

    return self.__loan_uninit()


def loan_slice_uninit(self: Publisher, number_of_elements: int) -> SampleMutUninit:
    """
    Loans/allocates a `SampleMutUninit` from the underlying data segment of the `Publisher`.

    The user has to initialize the payload before it can be sent.
    Fails when it is called for data types which are not a slice.
    On failure it returns `LoanError` describing the failure.
    """
    assert get_origin(self.__payload_type_details) is Slice

    return self.__loan_slice_uninit(number_of_elements)


def initial_max_slice_len(
    self: PortFactoryPublisher, value: int
) -> PortFactoryPublisher:
    """Sets the maximum slice length that a user can allocate."""
    assert get_origin(self.__payload_type_details) is Slice

    return self.__initial_max_slice_len(value)


def allocation_strategy(
    self: PortFactoryPublisher, value: AllocationStrategy
) -> PortFactoryPublisher:
    """Defines the allocation strategy that is used when the memory is exhausted."""
    assert get_origin(self.__payload_type_details) is Slice

    return self.__allocation_strategy(value)


PortFactoryPublisher.initial_max_slice_len = initial_max_slice_len
PortFactoryPublisher.allocation_strategy = allocation_strategy

Publisher.send_copy = send_copy
Publisher.loan_uninit = loan_uninit
Publisher.loan_slice_uninit = loan_slice_uninit

Sample.payload = payload
Sample.user_header = user_header

SampleMut.payload = payload
SampleMut.user_header = user_header

SampleMutUninit.write_payload = write_payload
SampleMutUninit.payload = payload
SampleMutUninit.user_header = user_header

ServiceBuilder.publish_subscribe = publish_subscribe
ServiceBuilderPublishSubscribe.user_header = set_user_header



# Copyright (c) 2025 Contributors to the Eclipse Foundation
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Apache Software License 2.0 which is available at
# https://www.apache.org/licenses/LICENSE-2.0, or the MIT license
# which is available at https://opensource.org/licenses/MIT.
#
# SPDX-License-Identifier: Apache-2.0 OR MIT

import iceoryx2 as iox2
import pytest

service_types = [iox2.ServiceType.Ipc, iox2.ServiceType.Local]


@pytest.mark.parametrize("service_type", service_types)
def test_creating_node_works(service_type: iox2.ServiceType) -> None:
    sut = iox2.NodeBuilder.new().create(service_type)
    assert sut.name.as_str() == ""


@pytest.mark.parametrize("service_type", service_types)
def test_creating_node_with_properties_works(
    service_type: iox2.ServiceType,
) -> None:
    node_name = iox2.testing.generate_node_name()
    signal_handling_mode = iox2.SignalHandlingMode.Disabled
    config = iox2.testing.generate_isolated_config()

    sut = (
        iox2.NodeBuilder.new()
        .name(node_name)
        .signal_handling_mode(signal_handling_mode)
        .config(config)
        .create(service_type)
    )
    assert sut.name == node_name
    assert sut.signal_handling_mode == signal_handling_mode
    assert sut.config == config


@pytest.mark.parametrize("service_type", service_types)
def test_cleanup_dead_nodes_can_be_called(
    service_type: iox2.ServiceType,
) -> None:
    try:
        iox2.Node.cleanup_dead_nodes(
            service_type, iox2.testing.generate_isolated_config()
        )
    except iox2.NodeCleanupFailure:
        assert False


@pytest.mark.parametrize("service_type", service_types)
def test_created_nodes_can_be_listed(service_type: iox2.ServiceType) -> None:
    config = iox2.testing.generate_isolated_config()
    sut_1 = (
        iox2.NodeBuilder.new()
        .name(iox2.NodeName.new("behind you, there is"))
        .config(config)
        .create(service_type)
    )
    sut_2 = (
        iox2.NodeBuilder.new()
        .name(iox2.NodeName.new("a 3 headed monkey"))
        .config(config)
        .create(service_type)
    )

    node_list = iox2.Node.list(service_type, config)
    assert len(node_list) == 2
    for node in node_list:
        match node:
            case node.Alive():
                assert isinstance(node, iox2.NodeState.Alive)
                assert node[0].id in (sut_1.id, sut_2.id)
                assert node[0].details.name in (sut_1.name, sut_2.name)
            case node.Dead():
                assert False
            case node.Inaccessible():
                assert False
            case node.Undefined():
                assert False


@pytest.mark.parametrize("service_type", service_types)
def test_wait_can_be_called(service_type: iox2.ServiceType) -> None:
    config = iox2.testing.generate_isolated_config()
    sut = iox2.NodeBuilder.new().config(config).create(service_type)
    try:
        sut.wait(iox2.Duration.from_millis(1))
    except iox2.NodeWaitFailure:
        assert False



# Copyright (c) 2025 Contributors to the Eclipse Foundation
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Apache Software License 2.0 which is available at
# https://www.apache.org/licenses/LICENSE-2.0, or the MIT license
# which is available at https://opensource.org/licenses/MIT.
#
# SPDX-License-Identifier: Apache-2.0 OR MIT

import ctypes

import iceoryx2 as iox2
import pytest

service_types = [iox2.ServiceType.Ipc, iox2.ServiceType.Local]


class Payload(ctypes.Structure):
    _fields_ = [("data", ctypes.c_ubyte)]


class HeaderPayload(ctypes.Structure):
    _fields_ = [("data", ctypes.c_ubyte), ("fuu", ctypes.c_int)]


@pytest.mark.parametrize("service_type", service_types)
def test_non_existing_service_can_be_created(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)
    try:
        service_name = iox2.testing.generate_service_name()
        sut = node.service_builder(service_name).publish_subscribe(Payload).create()
        assert sut.name == service_name
    except iox2.PublishSubscribeCreateError:
        assert False


@pytest.mark.parametrize("service_type", service_types)
def test_existing_service_cannot_be_created(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    _existing_service = (
        node.service_builder(service_name).publish_subscribe(Payload).create()
    )

    with pytest.raises(iox2.PublishSubscribeCreateError):
        node.service_builder(service_name).publish_subscribe(Payload).create()


@pytest.mark.parametrize("service_type", service_types)
def test_existing_service_can_be_opened(service_type: iox2.ServiceType) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    _existing_service = (
        node.service_builder(service_name).publish_subscribe(Payload).create()
    )
    try:
        sut = node.service_builder(service_name).publish_subscribe(Payload).open()
        assert sut.name == service_name
    except iox2.PublishSubscribeOpenError:
        assert False


@pytest.mark.parametrize("service_type", service_types)
def test_non_existing_service_cannot_be_opened(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    with pytest.raises(iox2.PublishSubscribeOpenError):
        node.service_builder(service_name).publish_subscribe(Payload).open()


@pytest.mark.parametrize("service_type", service_types)
def test_non_existing_service_is_created_with_open_or_create(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    try:
        sut = (
            node.service_builder(service_name)
            .publish_subscribe(Payload)
            .open_or_create()
        )
        assert sut.name == service_name
    except iox2.PublishSubscribeOpenOrCreateError:
        assert False


@pytest.mark.parametrize("service_type", service_types)
def test_existing_service_is_opened_with_open_or_create(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    _existing_service = (
        node.service_builder(service_name).publish_subscribe(Payload).create()
    )

    try:
        sut = (
            node.service_builder(service_name)
            .publish_subscribe(Payload)
            .open_or_create()
        )
        assert sut.name == service_name
    except iox2.PublishSubscribeOpenOrCreateError:
        assert False


@pytest.mark.parametrize("service_type", service_types)
def test_create_service_with_attributes_work(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)
    attribute_spec = iox2.AttributeSpecifier.new().define(
        iox2.AttributeKey.new("fuu"), iox2.AttributeValue.new("bar")
    )

    service_name = iox2.testing.generate_service_name()
    sut_create = (
        node.service_builder(service_name)
        .publish_subscribe(Payload)
        .create_with_attributes(attribute_spec)
    )

    sut_open = node.service_builder(service_name).publish_subscribe(Payload).open()

    assert sut_create.attributes == attribute_spec.attributes
    assert sut_create.attributes == sut_open.attributes


@pytest.mark.parametrize("service_type", service_types)
def test_open_or_create_service_with_attributes_work(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    attribute_spec = iox2.AttributeSpecifier.new().define(
        iox2.AttributeKey.new("what"), iox2.AttributeValue.new("ever")
    )
    attribute_verifier = iox2.AttributeVerifier.new().require(
        iox2.AttributeKey.new("what"), iox2.AttributeValue.new("ever")
    )

    service_name = iox2.testing.generate_service_name()
    sut_create = (
        node.service_builder(service_name)
        .publish_subscribe(Payload)
        .open_or_create_with_attributes(attribute_verifier)
    )

    sut_open = node.service_builder(service_name).publish_subscribe(Payload).open()

    assert sut_create.attributes == attribute_spec.attributes
    assert sut_create.attributes == sut_open.attributes


@pytest.mark.parametrize("service_type", service_types)
def test_open_service_with_attributes_work(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    attribute_spec = iox2.AttributeSpecifier.new().define(
        iox2.AttributeKey.new("knock"), iox2.AttributeValue.new("knock")
    )
    attribute_verifier = iox2.AttributeVerifier.new().require(
        iox2.AttributeKey.new("knock"), iox2.AttributeValue.new("knock")
    )

    service_name = iox2.testing.generate_service_name()
    sut_create = (
        node.service_builder(service_name)
        .publish_subscribe(Payload)
        .create_with_attributes(attribute_spec)
    )

    sut_open = (
        node.service_builder(service_name)
        .publish_subscribe(Payload)
        .open_with_attributes(attribute_verifier)
    )

    assert sut_create.attributes == attribute_spec.attributes
    assert sut_create.attributes == sut_open.attributes


@pytest.mark.parametrize("service_type", service_types)
def test_node_listing_works(service_type: iox2.ServiceType) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    sut = node.service_builder(service_name).publish_subscribe(Payload).create()

    nodes = sut.nodes

    assert len(nodes) == 1
    for n in nodes:
        match n:
            case n.Alive():
                assert n[0].id == node.id
            case node.Dead():
                assert False
            case node.Inaccessible():
                assert False
            case node.Undefined():
                assert False


@pytest.mark.parametrize("service_type", service_types)
def test_service_builder_configuration_works(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    safe_overflow = False
    subscriber_max_borrowed_samples = 10
    history_size = 29
    subscriber_max_buffer_size = 38
    max_subscribers = 47
    max_publishers = 56
    max_nodes = 65
    sut = (
        node.service_builder(service_name)
        .publish_subscribe(Payload)
        .enable_safe_overflow(safe_overflow)
        .subscriber_max_borrowed_samples(subscriber_max_borrowed_samples)
        .history_size(history_size)
        .subscriber_max_buffer_size(subscriber_max_buffer_size)
        .max_subscribers(max_subscribers)
        .max_publishers(56)
        .max_nodes(65)
        .create()
    )

    static_config = sut.static_config
    assert static_config.max_nodes == max_nodes
    assert static_config.max_publishers == max_publishers
    assert static_config.max_subscribers == max_subscribers
    assert static_config.history_size == history_size
    assert static_config.subscriber_max_buffer_size == subscriber_max_buffer_size
    assert (
        static_config.subscriber_max_borrowed_samples == subscriber_max_borrowed_samples
    )
    assert static_config.has_safe_overflow == safe_overflow


@pytest.mark.parametrize("service_type", service_types)
def test_service_builder_based_on_custom_config_works(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    max_nodes = 112
    config.defaults.publish_subscribe.max_nodes = max_nodes
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    sut = node.service_builder(service_name).publish_subscribe(Payload).create()

    static_config = sut.static_config
    assert static_config.max_nodes == max_nodes


@pytest.mark.parametrize("service_type", service_types)
def test_custom_user_header_works(service_type: iox2.ServiceType) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    user_header = (
        iox2.TypeDetail.new()
        .type_variant(iox2.TypeVariant.FixedSize)
        .type_name(iox2.TypeName.new("HeaderPayload"))
        .size(ctypes.sizeof(HeaderPayload))
        .alignment(ctypes.alignment(HeaderPayload))
    )
    sut = (
        node.service_builder(service_name)
        .publish_subscribe(Payload)
        .user_header(HeaderPayload)
        .create()
    )

    assert sut.static_config.message_type_details.user_header == user_header


@pytest.mark.parametrize("service_type", service_types)
def test_custom_payload_works(service_type: iox2.ServiceType) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)

    service_name = iox2.testing.generate_service_name()
    payload = (
        iox2.TypeDetail.new()
        .type_variant(iox2.TypeVariant.FixedSize)
        .type_name(iox2.TypeName.new("Payload"))
        .size(ctypes.sizeof(Payload))
        .alignment(ctypes.alignment(Payload))
    )
    sut = node.service_builder(service_name).publish_subscribe(Payload).create()

    assert sut.static_config.message_type_details.payload == payload



# Copyright (c) 2025 Contributors to the Eclipse Foundation
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Apache Software License 2.0 which is available at
# https://www.apache.org/licenses/LICENSE-2.0, or the MIT license
# which is available at https://opensource.org/licenses/MIT.
#
# SPDX-License-Identifier: Apache-2.0 OR MIT

import iceoryx2 as iox2
import pytest

service_types = [iox2.ServiceType.Ipc, iox2.ServiceType.Local]


@pytest.mark.parametrize("service_type", service_types)
def test_newly_created_waitset_is_empty(
    service_type: iox2.ServiceType,
) -> None:
    sut = iox2.WaitSetBuilder.new().create(service_type)
    assert sut.len == 0
    assert sut.is_empty


@pytest.mark.parametrize("service_type", service_types)
def test_waitset_builder_set_signal_handling_mode_correctly(
    service_type: iox2.ServiceType,
) -> None:
    sut_1 = (
        iox2.WaitSetBuilder.new()
        .signal_handling_mode(iox2.SignalHandlingMode.Disabled)
        .create(service_type)
    )
    sut_2 = (
        iox2.WaitSetBuilder.new()
        .signal_handling_mode(iox2.SignalHandlingMode.HandleTerminationRequests)
        .create(service_type)
    )
    assert sut_1.signal_handling_mode == iox2.SignalHandlingMode.Disabled
    assert (
        sut_2.signal_handling_mode == iox2.SignalHandlingMode.HandleTerminationRequests
    )


@pytest.mark.parametrize("service_type", service_types)
def test_attaching_notifications_works(
    service_type: iox2.ServiceType,
) -> None:
    number_of_attachments = 15
    config = iox2.testing.generate_isolated_config()
    service_name = iox2.testing.generate_service_name()
    node = iox2.NodeBuilder.new().config(config).create(service_type)
    service = (
        node.service_builder(service_name)
        .event()
        .max_listeners(number_of_attachments)
        .create()
    )
    listeners = []
    for i in range(0, number_of_attachments):
        listener = service.listener_builder().create()
        listeners.append(listener)

    sut = iox2.WaitSetBuilder.new().create(service_type)
    waitset_guards = []
    for i in range(0, number_of_attachments):
        guard = sut.attach_notification(listeners[i])
        waitset_guards.append(guard)
        assert sut.len == i + 1
        assert not sut.is_empty


@pytest.mark.parametrize("service_type", service_types)
def test_attaching_deadlines_works(
    service_type: iox2.ServiceType,
) -> None:
    number_of_attachments = 15
    deadline = iox2.Duration.from_millis(123)
    config = iox2.testing.generate_isolated_config()
    service_name = iox2.testing.generate_service_name()
    node = iox2.NodeBuilder.new().config(config).create(service_type)
    service = (
        node.service_builder(service_name)
        .event()
        .max_listeners(number_of_attachments)
        .create()
    )
    listeners = []
    for i in range(0, number_of_attachments):
        listener = service.listener_builder().create()
        listeners.append(listener)

    sut = iox2.WaitSetBuilder.new().create(service_type)
    waitset_guards = []
    for i in range(0, number_of_attachments):
        guard = sut.attach_deadline(listeners[i], deadline)
        waitset_guards.append(guard)
        assert sut.len == i + 1
        assert not sut.is_empty


@pytest.mark.parametrize("service_type", service_types)
def test_attaching_interval_works(
    service_type: iox2.ServiceType,
) -> None:
    number_of_attachments = 15

    sut = iox2.WaitSetBuilder.new().create(service_type)
    waitset_guards = []
    for i in range(0, number_of_attachments):
        guard = sut.attach_interval(iox2.Duration.from_millis(i + 1))
        waitset_guards.append(guard)
        assert sut.len == i + 1
        assert not sut.is_empty


@pytest.mark.parametrize("service_type", service_types)
def test_wait_and_process_returns_when_timeout_has_passed(
    service_type: iox2.ServiceType,
) -> None:
    sut = iox2.WaitSetBuilder.new().create(service_type)
    _guard = sut.attach_interval(iox2.Duration.from_secs(123))
    (triggers, result) = sut.wait_and_process_with_timeout(iox2.Duration.from_millis(1))
    assert len(triggers) == 0
    assert result == iox2.WaitSetRunResult.AllEventsHandled


@pytest.mark.parametrize("service_type", service_types)
def test_wait_and_process_returns_triggered_listeners(
    service_type: iox2.ServiceType,
) -> None:
    number_of_attachments = 15
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)
    services = []
    listeners = []
    notifiers = []
    for i in range(0, number_of_attachments):
        service_name = iox2.testing.generate_service_name()
        service = (
            node.service_builder(service_name)
            .event()
            .max_listeners(number_of_attachments)
            .create()
        )
        listener = service.listener_builder().create()
        notifier = service.notifier_builder().create()
        listeners.append(listener)
        notifiers.append(notifier)
        services.append(service)

    sut = iox2.WaitSetBuilder.new().create(service_type)
    waitset_guards = []
    for i in range(0, number_of_attachments):
        guard = sut.attach_notification(listeners[i])
        waitset_guards.append(guard)
        assert sut.len == i + 1
        assert not sut.is_empty

    notifiers[0].notify()
    notifiers[1].notify()

    (triggers, result) = sut.wait_and_process()
    assert len(triggers) == 2
    assert result == iox2.WaitSetRunResult.AllEventsHandled

    for i in range(0, 1):
        assert triggers[i].has_event_from(waitset_guards[0]) or triggers[
            i
        ].has_event_from(waitset_guards[1])

    for k in range(2, number_of_attachments):
        for i in range(0, 1):
            assert not triggers[i].has_event_from(waitset_guards[k])


@pytest.mark.parametrize("service_type", service_types)
def test_reports_missed_deadline(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)
    service_name = iox2.testing.generate_service_name()
    service = node.service_builder(service_name).event().create()
    listener = service.listener_builder().create()

    sut = iox2.WaitSetBuilder.new().create(service_type)
    guard = sut.attach_deadline(listener, iox2.Duration.from_nanos(1))

    (triggers, result) = sut.wait_and_process()
    assert len(triggers) == 1
    assert result == iox2.WaitSetRunResult.AllEventsHandled

    assert triggers[0].has_missed_deadline(guard)


@pytest.mark.parametrize("service_type", service_types)
def test_create_attachment_id_from_guard(
    service_type: iox2.ServiceType,
) -> None:
    config = iox2.testing.generate_isolated_config()
    node = iox2.NodeBuilder.new().config(config).create(service_type)
    service_name = iox2.testing.generate_service_name()
    service = node.service_builder(service_name).event().create()
    listener = service.listener_builder().create()

    sut = iox2.WaitSetBuilder.new().create(service_type)
    guard = sut.attach_deadline(listener, iox2.Duration.from_nanos(1))
    attachment_id = iox2.WaitSetAttachmentId.from_guard(guard)

    (triggers, result) = sut.wait_and_process()
    assert len(triggers) == 1
    assert result == iox2.WaitSetRunResult.AllEventsHandled

    assert triggers[0] == attachment_id


@pytest.mark.parametrize("service_type", service_types)
def test_deleting_guard_explicitly_removes_attachment(
    service_type: iox2.ServiceType,
) -> None:
    number_of_attachments = 15
    config = iox2.testing.generate_isolated_config()
    service_name = iox2.testing.generate_service_name()
    node = iox2.NodeBuilder.new().config(config).create(service_type)
    service = (
        node.service_builder(service_name)
        .event()
        .max_listeners(number_of_attachments)
        .create()
    )
    listeners = []
    for i in range(0, number_of_attachments):
        listener = service.listener_builder().create()
        listeners.append(listener)

    sut = iox2.WaitSetBuilder.new().create(service_type)
    waitset_guards = []
    for i in range(0, number_of_attachments):
        guard = sut.attach_notification(listeners[i])
        waitset_guards.append(guard)

    for i in range(0, number_of_attachments):
        assert sut.len == number_of_attachments - i
        waitset_guards[i].delete()

    assert sut.len == 0
    assert sut.is_empty



# Copyright (c) 2025 Contributors to the Eclipse Foundation
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Apache Software License 2.0 which is available at
# https://www.apache.org/licenses/LICENSE-2.0, or the MIT license
# which is available at https://opensource.org/licenses/MIT.
#
# SPDX-License-Identifier: Apache-2.0 OR MIT

import ctypes

import iceoryx2 as iox2
import pytest


def test_slice_contains_correct_number_of_elements() -> None:
    data_ptr = 1234
    number_of_elements = 78
    sut = iox2.Slice[ctypes.c_uint8](data_ptr, number_of_elements, ctypes.c_uint8)

    assert sut.len() == number_of_elements


def test_slice_returns_correct_data_ptr() -> None:
    data_ptr = 123
    number_of_elements = 7809
    sut = iox2.Slice[ctypes.c_uint8](data_ptr, number_of_elements, ctypes.c_uint8)

    assert sut.as_ptr() == data_ptr


def test_get_set_item_raises_exception_when_out_of_bounds() -> None:
    data_ptr = 23
    number_of_elements = 5

    sut = iox2.Slice[ctypes.c_uint8](data_ptr, number_of_elements, ctypes.c_uint8)

    with pytest.raises(IndexError):
        _unused = sut[number_of_elements + 1]

    with pytest.raises(IndexError):
        sut[number_of_elements + 1] = ctypes.c_uint8(7)

    with pytest.raises(IndexError):
        _unused = sut[-1]

    with pytest.raises(IndexError):
        sut[-1] = ctypes.c_uint8(7)



# Copyright (c) 2025 Contributors to the Eclipse Foundation
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Apache Software License 2.0 which is available at
# https://www.apache.org/licenses/LICENSE-2.0, or the MIT license
# which is available at https://opensource.org/licenses/MIT.
#
# SPDX-License-Identifier: Apache-2.0 OR MIT

"""Strong type safe extensions for the request-response messaging pattern."""

import ctypes
from typing import Any, Type, TypeVar, get_args, get_origin

from ._iceoryx2 import *
from .slice import Slice
from .type_name import get_type_name

ReqT = TypeVar("ReqT", bound=ctypes.Structure)
ResT = TypeVar("ResT", bound=ctypes.Structure)


def request_response(
    self: ServiceBuilder, request: Type[ReqT], response: Type[ResT]
) -> ServiceBuilderPublishSubscribe:
    """
    Returns the `ServiceBuilderRequestResponse` to create a new request-response service.

    The request/response payload ctype must be provided as argument.
    """
    request_type_name = request.__name__
    request_type_size = 0
    request_type_align = 0
    request_type_variant = TypeVariant.FixedSize
    response_type_name = response.__name__
    response_type_size = 0
    response_type_align = 0
    response_type_variant = TypeVariant.FixedSize

    if get_origin(request) is Slice:
        (contained_type,) = get_args(request)
        request_type_name = get_type_name(contained_type)
        request_type_variant = TypeVariant.Dynamic
        request_type_size = ctypes.sizeof(contained_type)
        request_type_align = ctypes.alignment(contained_type)
    else:
        request_type_name = get_type_name(request)
        request_type_size = ctypes.sizeof(request)
        request_type_align = ctypes.alignment(request)
        request_type_variant = TypeVariant.FixedSize

    if get_origin(response) is Slice:
        (contained_type,) = get_args(response)
        response_type_name = get_type_name(contained_type)
        response_type_variant = TypeVariant.Dynamic
        response_type_size = ctypes.sizeof(contained_type)
        response_type_align = ctypes.alignment(contained_type)
    else:
        response_type_name = get_type_name(response)
        response_type_size = ctypes.sizeof(response)
        response_type_align = ctypes.alignment(response)
        response_type_variant = TypeVariant.FixedSize

    result = self.__request_response()
    result.__set_request_payload_type(request)
    result.__set_response_payload_type(response)

    return (
        result.__request_payload_type_details(
            TypeDetail.new()
            .type_variant(request_type_variant)
            .type_name(TypeName.new(request_type_name))
            .size(request_type_size)
            .alignment(request_type_align)
        )
        .__request_header_type_details(
            TypeDetail.new()
            .type_variant(TypeVariant.FixedSize)
            .type_name(TypeName.new("()"))
            .size(0)
            .alignment(1)
        )
        .__response_payload_type_details(
            TypeDetail.new()
            .type_variant(response_type_variant)
            .type_name(TypeName.new(response_type_name))
            .size(response_type_size)
            .alignment(response_type_align)
        )
        .__response_header_type_details(
            TypeDetail.new()
            .type_variant(TypeVariant.FixedSize)
            .type_name(TypeName.new("()"))
            .size(0)
            .alignment(1)
        )
    )


def set_request_header(
    self: ServiceBuilderPublishSubscribe, request: Type[ReqT]
) -> ServiceBuilderPublishSubscribe:
    """Sets the request header type for the service."""
    type_name = get_type_name(request)
    result = self.__request_header_type_details(
        TypeDetail.new()
        .type_variant(TypeVariant.FixedSize)
        .type_name(TypeName.new(type_name))
        .size(ctypes.sizeof(request))
        .alignment(ctypes.alignment(request))
    )
    result.__set_request_header_type(request)
    return result


def set_response_header(
    self: ServiceBuilderPublishSubscribe, response: Type[ResT]
) -> ServiceBuilderPublishSubscribe:
    """Sets the response header type for the service."""
    type_name = get_type_name(response)
    result = self.__response_header_type_details(
        TypeDetail.new()
        .type_variant(TypeVariant.FixedSize)
        .type_name(TypeName.new(type_name))
        .size(ctypes.sizeof(response))
        .alignment(ctypes.alignment(response))
    )
    result.__set_response_header_type(response)
    return result


def request_payload(self: Any) -> Any:
    """Returns a `ctypes.POINTER` to the requests payload."""
    assert self.__request_payload_type_details is not None
    if get_origin(self.__request_payload_type_details) is Slice:
        (contained_type,) = get_args(self.__request_payload_type_details)
        return Slice(self.payload_ptr, self.__slice_len, contained_type)

    return ctypes.cast(
        self.payload_ptr, ctypes.POINTER(self.__request_payload_type_details)
    )


def response_payload(self: Any) -> Any:
    """Returns a `ctypes.POINTER` to the responses payload."""
    assert self.__response_payload_type_details is not None
    if get_origin(self.__response_payload_type_details) is Slice:
        (contained_type,) = get_args(self.__response_payload_type_details)
        return Slice(self.payload_ptr, self.__slice_len, contained_type)

    return ctypes.cast(
        self.payload_ptr, ctypes.POINTER(self.__response_payload_type_details)
    )


def request_header(self: Any) -> Any:
    """Returns a `ctypes.POINTER` to the request header."""
    assert self.__request_header_type_details is not None
    return ctypes.cast(
        self.user_header_ptr, ctypes.POINTER(self.__request_header_type_details)
    )


def response_header(self: Any) -> Any:
    """Returns a `ctypes.POINTER` to the response header."""
    assert self.__response_header_type_details is not None
    return ctypes.cast(
        self.user_header_ptr,
        ctypes.POINTER(self.__response_header_type_details),
    )


def write_request_payload(self: RequestMutUninit, t: Type[ReqT]) -> RequestMut:
    """Writes the provided payload into the request."""
    assert self.__request_payload_type_details is not None
    assert ctypes.sizeof(t) == ctypes.sizeof(self.__request_payload_type_details)
    assert ctypes.alignment(t) == ctypes.alignment(self.__request_payload_type_details)

    ctypes.memmove(self.payload_ptr, ctypes.byref(t), ctypes.sizeof(t))
    return self.assume_init()


def write_response_payload(self: ResponseMutUninit, t: Type[ReqT]) -> ResponseMut:
    """Writes the provided payload into the response."""
    assert self.__response_payload_type_details is not None
    assert ctypes.sizeof(t) == ctypes.sizeof(self.__response_payload_type_details)
    assert ctypes.alignment(t) == ctypes.alignment(self.__response_payload_type_details)

    ctypes.memmove(self.payload_ptr, ctypes.byref(t), ctypes.sizeof(t))
    return self.assume_init()


def loan_uninit_request(self: Client) -> RequestMutUninit:
    """
    Loans/allocates memory from the underlying data segment.

    The user has to initialize the payload before it can be sent. On failure it returns
    `LoanError` describing the failure.
    """
    assert not get_origin(self.__request_payload_type_details) is Slice

    return self.__loan_uninit()


def loan_uninit_response(self: ActiveRequest) -> ResponseMutUninit:
    """
    Loans/allocates memory from the underlying data segment.

    The user has to initialize the payload before it can be sent. On failure it returns
    `LoanError` describing the failure.
    """
    assert not get_origin(self.__response_payload_type_details) is Slice

    return self.__loan_uninit()


def loan_slice_uninit_request(
    self: Client, number_of_elements: int
) -> RequestMutUninit:
    """
    Loans/allocates memory from the underlying data segment.

    The user has to initialize the payload before it can be sent.
    Fails when it is called for data types which are not a slice.
    On failure it returns `LoanError` describing the failure.
    """
    assert get_origin(self.__request_payload_type_details) is Slice

    return self.__loan_slice_uninit(number_of_elements)


def loan_slice_uninit_response(
    self: ActiveRequest, number_of_elements: int
) -> ResponseMutUninit:
    """
    Loans/allocates memory from the underlying data segment.

    The user has to initialize the payload before it can be sent.
    Fails when it is called for data types which are not a slice.
    On failure it returns `LoanError` describing the failure.
    """
    assert get_origin(self.__response_payload_type_details) is Slice

    return self.__loan_slice_uninit(number_of_elements)


def initial_max_slice_len_request(
    self: PortFactoryClient, value: int
) -> PortFactoryClient:
    """Sets the maximum slice length that a user can allocate."""
    assert get_origin(self.__request_payload_type_details) is Slice

    return self.__initial_max_slice_len(value)


def initial_max_slice_len_response(
    self: PortFactoryServer, value: int
) -> PortFactoryServer:
    """Sets the maximum slice length that a user can allocate."""
    assert get_origin(self.__response_payload_type_details) is Slice

    return self.__initial_max_slice_len(value)


def allocation_strategy_request(
    self: PortFactoryClient, value: AllocationStrategy
) -> PortFactoryClient:
    """Defines the allocation strategy that is used when the memory is exhausted."""
    assert get_origin(self.__request_payload_type_details) is Slice

    return self.__allocation_strategy(value)


def allocation_strategy_response(
    self: PortFactoryServer, value: AllocationStrategy
) -> PortFactoryServer:
    """Defines the allocation strategy that is used when the memory is exhausted."""
    assert get_origin(self.__response_payload_type_details) is Slice

    return self.__allocation_strategy(value)


def send_request_copy(self: Client, t: Type[ReqT]) -> PendingResponse:
    """Sends a copy of the provided type."""
    assert self.__request_payload_type_details is not None
    request_uninit = self.__loan_uninit()

    assert ctypes.sizeof(t) == ctypes.sizeof(
        request_uninit.__request_payload_type_details
    )
    assert ctypes.alignment(t) == ctypes.alignment(
        request_uninit.__request_payload_type_details
    )

    ctypes.memmove(request_uninit.payload_ptr, ctypes.byref(t), ctypes.sizeof(t))
    request = request_uninit.assume_init()
    return request.send()


def send_response_copy(self: ActiveRequest, t: Type[ResT]) -> Any:
    """Sends a copy of the provided type."""
    assert self.__response_payload_type_details is not None
    response_uninit = self.__loan_uninit()

    assert ctypes.sizeof(t) == ctypes.sizeof(
        response_uninit.__response_payload_type_details
    )
    assert ctypes.alignment(t) == ctypes.alignment(
        response_uninit.__response_payload_type_details
    )

    ctypes.memmove(response_uninit.payload_ptr, ctypes.byref(t), ctypes.sizeof(t))
    response = response_uninit.assume_init()
    return response.send()


ServiceBuilder.request_response = request_response
ServiceBuilderRequestResponse.request_header = set_request_header
ServiceBuilderRequestResponse.response_header = set_response_header

ActiveRequest.send_copy = send_response_copy
ActiveRequest.payload = request_payload
ActiveRequest.user_header = request_header
ActiveRequest.loan_uninit = loan_uninit_response
ActiveRequest.loan_slice_uninit = loan_slice_uninit_response

PortFactoryClient.initial_max_slice_len = initial_max_slice_len_request
PortFactoryClient.allocation_strategy = allocation_strategy_request
PortFactoryServer.initial_max_slice_len = initial_max_slice_len_response
PortFactoryServer.allocation_strategy = allocation_strategy_response

PendingResponse.payload = request_payload
PendingResponse.user_header = request_header

RequestMut.payload = request_payload
RequestMut.user_header = request_header

RequestMutUninit.payload = request_payload
RequestMutUninit.user_header = request_header
RequestMutUninit.write_payload = write_request_payload

Response.payload = response_payload
Response.user_header = response_header

ResponseMut.payload = response_payload
ResponseMut.user_header = response_header

ResponseMutUninit.payload = response_payload
ResponseMutUninit.user_header = response_header
ResponseMutUninit.write_payload = write_response_payload

Client.loan_uninit = loan_uninit_request
Client.loan_slice_uninit = loan_slice_uninit_request
Client.send_copy = send_request_copy



# Copyright (c) 2025 Contributors to the Eclipse Foundation
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Apache Software License 2.0 which is available at
# https://www.apache.org/licenses/LICENSE-2.0, or the MIT license
# which is available at https://opensource.org/licenses/MIT.
#
# SPDX-License-Identifier: Apache-2.0 OR MIT

"""Generates a human readable type name from a given type."""

import ctypes
from typing import Any, Type, TypeVar

T = TypeVar("T")


def get_type_name(t: Type[T]) -> Any:
    """Generates a human readable type name from a given type."""
    if hasattr(t, "type_name"):
        return t.type_name()

    if t.__name__ == "c_ubyte" and ctypes.sizeof(t) == 1:
        return "u8"
    if t.__name__ == "c_ushort" and ctypes.sizeof(t) == 2:
        return "u16"
    if t.__name__ == "c_uint" and ctypes.sizeof(t) == 4:
        return "u32"
    if t.__name__ == "c_ulong" and ctypes.sizeof(t) == 8:
        return "u64"
    if t.__name__ == "c_byte" and ctypes.sizeof(t) == 1:
        return "i8"
    if t.__name__ == "c_short" and ctypes.sizeof(t) == 2:
        return "i16"
    if t.__name__ == "c_int" and ctypes.sizeof(t) == 4:
        return "i32"
    if t.__name__ == "c_long" and ctypes.sizeof(t) == 8:
        return "i64"
    if t.__name__ == "c_bool":
        return "bool"
    if t.__name__ == "c_float" and ctypes.sizeof(t) == 4:
        return "f32"
    if t.__name__ == "c_double" and ctypes.sizeof(t) == 8:
        return "f64"

    return t.__name__



# Copyright (c) 2025 Contributors to the Eclipse Foundation
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Apache Software License 2.0 which is available at
# https://www.apache.org/licenses/LICENSE-2.0, or the MIT license
# which is available at https://opensource.org/licenses/MIT.
#
# SPDX-License-Identifier: Apache-2.0 OR MIT

import iceoryx2 as iox2
import pytest


def test_file_path_can_be_constructed() -> None:
    sut = iox2.FilePath.new("some/file")
    assert sut.to_string() == "some/file"


def test_file_path_with_invalid_content_cannot_be_constructed() -> None:
    invalid_content = "*/i/am/not/a/path*"
    with pytest.raises(iox2.SemanticStringError):
        iox2.FilePath.new(invalid_content)
